"""cubicweb-eac application package

Implementation of Encoded Archival Context for CubicWeb
"""

from functools import partial
from cubicweb_compound import CompositeGraph


# EAC mappings

TYPE_MAPPING = {
    "corporateBody": "authority",
    "person": "person",
    "family": "family",
}

MAINTENANCETYPE_MAPPING = {
    "created": "create",
    "revised": "modify",
}

# Order matters for this one in order to export correctly
ADDRESS_MAPPING = [
    ("voie", "street"),
    ("postalcode", "postalcode"),
    ("city", "city"),
]


AuthorityRecordGraph = partial(CompositeGraph, skiprtypes=("generated", "used"))



# copyright 2015-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac common test tools"""

from os.path import basename

from doctest import Example

from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker

from cubicweb.dataimport.importer import SimpleImportLog


def authority_record(cnx, record_id, name, kind="person", **kwargs):
    """Return an AuthorityRecord with specified kind and name."""
    kind_eid = cnx.find("AgentKind", name=kind)[0][0]
    if "reverse_name_entry_for" not in kwargs:
        kwargs["reverse_name_entry_for"] = cnx.create_entity(
            "NameEntry", parts=name, form_variant="authorized"
        )
    return cnx.create_entity(
        "AuthorityRecord", record_id=record_id, agent_kind=kind_eid, **kwargs
    )


def eac_import(cnx, fpath):
    import_log = SimpleImportLog(basename(fpath))
    created, updated, _ = cnx.call_service(
        "eac.import", stream=fpath, import_log=import_log, raise_on_error=True
    )
    return created, updated


class XmlTestMixin(object):
    """Mixin class provinding additional assertion methods for checking XML data."""

    def assertXmlEqual(self, actual, expected):
        """Check that both XML strings represent the same XML tree."""
        checker = LXMLOutputChecker()
        if not checker.check_output(expected, actual, 0):
            message = checker.output_difference(Example("", expected), actual, 0)
            self.fail(message)

    def assertXmlValid(self, xml_data, xsd_filename, debug=False):
        """Validate an XML file (.xml) according to an XML schema (.xsd)."""
        with open(xsd_filename) as xsd:
            xmlschema = etree.XMLSchema(etree.parse(xsd))
        # Pretty-print xml_data to get meaningful line information.
        xml_data = etree.tostring(etree.fromstring(xml_data), pretty_print=True)
        xml_data = etree.fromstring(xml_data)
        if debug and not xmlschema.validate(xml_data):
            print(etree.tostring(xml_data, pretty_print=True))
        xmlschema.assertValid(xml_data)



# copyright 2016-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac specific hooks and operations"""

from cubicweb import (
    _,
    ValidationError,
)
from cubicweb.predicates import is_instance
from cubicweb.server import hook


class DeleteTernaryRelationsHook(hook.Hook):
    """When an AuthorityRecord entity is being deleted we need to delete all
    "qualified" (ternary) relations with another AuthorityRecord, otherwise
    we'll get a validation error about one side of the ternary relation being
    missing.
    """

    __select__ = hook.Hook.__select__ & is_instance("AuthorityRecord")
    __regid__ = "eac.delete-ternary-relations"
    events = ("before_delete_entity",)

    def __call__(self):
        errors, msgargs = {}, {}
        for rtype in (
            "association_from",
            "association_to",
            "chronological_predecessor",
            "chronological_successor",
            "hierarchical_parent",
            "hierarchical_child",
            "family_from",
            "family_to",
            "identity_from",
            "identity_to",
        ):
            for relation in self.entity.related(rtype, role="object").entities():
                if self._cw.deleted_in_transaction(relation.eid):
                    continue
                if relation.cw_has_perm("delete"):
                    self.info(
                        'deleting "%s" as %s-object is being deleted',
                        relation.dc_title(),
                        rtype,
                    )
                    relation.cw_delete()
                else:
                    argname = "{}-subject-{}".format(rtype, relation.eid)
                    errors[rtype] = _(
                        '"%({})s" would need to be deleted alongside '
                        "the AuthorityRecord but this is disallowed"
                    ).format(argname)
                    msgargs[argname] = relation.dc_title()
        if errors:
            raise ValidationError(self.entity.eid, errors, msgargs)


# ensure that equivalent_concept Concept has vocabulary_source defined ####################


class EnsureVocabularySource(hook.Hook):
    """When a equivalent_concept relation is set and targets a Concept, ensure that the
    vocabulary_source relation is set to the concept's scheme. This should not be necessary when
    using the UI where the workflow enforce setting the scheme first, but it's necessary during
    e.g. EAC import.
    """

    __select__ = hook.Hook.__select__ & hook.match_rtype(
        "equivalent_concept", toetypes=("Concept",)
    )
    __regid__ = "eac.add_equivalent_concept"
    events = ("after_add_relation",)

    def __call__(self):
        EnsureVocabularySourceOp.get_instance(self._cw).add_data(
            (self.eidfrom, self.eidto)
        )


class EnsureVocabularySourceOp(hook.DataOperationMixIn, hook.Operation):
    """Ensure X equivalent_concept target Concept as proper X vocabulary_source Scheme"""

    def precommit_event(self):
        cnx = self.cnx
        for eid, concept_eid in self.get_data():
            cnx.execute(
                "SET X vocabulary_source SC WHERE NOT X vocabulary_source SC, "
                "C in_scheme SC, C eid %(c)s, X eid %(x)s",
                {"x": eid, "c": concept_eid},
            )


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)

    # Add relations involved in a composite graph with security setup to "on
    # commit" check step.
    from cubicweb.server import ON_COMMIT_ADD_RELATIONS
    from cubicweb_compound.utils import mandatory_rdefs
    from cubicweb_eac import AuthorityRecordGraph

    graph = AuthorityRecordGraph(vreg.schema)
    for rdef, __ in mandatory_rdefs(
        vreg.schema, graph.parent_structure("AuthorityRecord")
    ):
        ON_COMMIT_ADD_RELATIONS.add(rdef.rtype)



# copyright 2015-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac dataimport utilities for EAC-CPF (Encoded Archival
Context -- Corporate Bodies, Persons, and Families).
"""

from collections import deque
import copy
import json
import datetime
import re
from functools import wraps, partial
import inspect
import logging
from uuid import uuid4

from dateutil.parser import parse as parse_date
from lxml import etree

from cubicweb.dataimport.importer import ExtEntity

from cubicweb_skos import to_unicode

from cubicweb_eac import TYPE_MAPPING, ADDRESS_MAPPING, MAINTENANCETYPE_MAPPING


TYPE_MAPPING = TYPE_MAPPING.copy()
TYPE_MAPPING["human"] = "person"

ETYPES_ORDER_HINT = (
    "AgentKind",
    "PhoneNumber",
    "PostalAddress",
    "AuthorityRecord",
    "Convention",
    "AgentPlace",
    "PlaceEntry",
    "Mandate",
    "LegalStatus",
    "History",
    "HistoricalEvent",
    "Structure",
    "AgentFunction",
    "Occupation",
    "GeneralContext",
    "AssociationRelation",
    "ChronologicalRelation",
    "HierarchicalRelation",
    "EACResourceRelation",
    "EACFunctionRelation",
    "ParallelNames",
    "ExternalUri",
    "EACSource",
    "Activity",
    "EACOtherRecordId",
    "NameEntry",
    "IdentityRelation",
    "FamilyRelation",
)


class InvalidEAC(RuntimeError):
    """EAC input is malformed."""


class InvalidXML(RuntimeError):
    """EAC input has an invalid XML format"""


class MissingTag(RuntimeError):
    """Mandatory tag is missing in EAC input"""

    def __init__(self, tag, tag_parent=None):
        super(MissingTag, self).__init__()
        self.tag = tag
        self.tag_parent = tag_parent


def external_uri(uri):
    values = [str(uri)]
    return ExtEntity("ExternalUri", uri, {"uri": set(values), "cwuri": set(values)})


def filter_none(func):
    """Filter None value from a generator function."""

    def wrapped(*args, **kwargs):
        for x in func(*args, **kwargs):
            if x is not None:
                yield x

    return wraps(func)(wrapped)


def filter_empty(func):
    """Filter out empty ExtEntity (i.e. with empty ``values`` attribute)."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        for extentity in func(self, *args, **kwargs):
            if extentity.values:
                yield extentity

    return wrapper


def elem_maybe_none(func):
    """Method decorator for external entity builder function handling the case
    of `elem` being None.
    """
    if inspect.isgeneratorfunction(func):

        def wrapped(self, elem, *args, **kwargs):
            if elem is None:
                return
            for extentity in func(self, elem, *args, **kwargs):
                yield extentity

    else:

        def wrapped(self, elem, *args, **kwargs):
            if elem is None:
                return None
            else:
                return func(self, elem, *args, **kwargs)

    return wraps(func)(wrapped)


def add_xml_wrap_for(*etypes):
    """Add an `xml_wrap` attribute in ExtEntity's values dictionnary."""

    def decorator(func):
        def wrapped(self, elem):
            objectXMLWrap = self._elem_find(elem, "eac:objectXMLWrap")
            xmlwrap = None
            if objectXMLWrap is not None:
                nchildren = len(objectXMLWrap)
                if nchildren >= 1:
                    xmlwrap = objectXMLWrap[0]
                if nchildren > 1:
                    msg = self._("multiple children elements found in {0}").format(
                        objectXMLWrap
                    )
                    self.import_log.record_warning(msg, line=objectXMLWrap.sourceline)
            attribute_added = False
            for extentity in func(self, elem):
                if xmlwrap is not None and extentity.etype in etypes:
                    # prevent association of xmlwrap to several extentities.
                    assert not attribute_added, "xml_wrap attribute already added"
                    extentity.values.setdefault("xml_wrap", set([])).add(
                        etree.tostring(xmlwrap, encoding="utf-8")
                    )
                    attribute_added = True
                yield extentity

        return wraps(func)(wrapped)

    return decorator


def add_items_for(etype):
    """Add an `items` attribute in ExtEntity's values dictionnary."""

    def decorator(func):
        @wraps(func)
        def wrapper(self, elem):
            for extentity in func(self, elem):
                if extentity.etype == etype:
                    items = self.parse_items(elem)
                    if items:
                        extentity.values["items"] = set([items])
                        extentity.values["items_format"] = set(["text/html"])
                yield extentity

        return wrapper

    return decorator


def relate_to_record_through(etype, rtype):
    """Add an ``rtype`` relationship from ``etype`` to the imported record."""

    def decorator(func):
        if inspect.isgeneratorfunction(func):

            def wrapper(self, *args, **kwargs):
                for extentity in func(self, *args, **kwargs):
                    if extentity.etype == etype:
                        extentity.values.setdefault(rtype, set()).add(self.record.extid)
                    yield extentity

        else:

            def wrapper(self, *args, **kwargs):
                extentity = func(self, *args, **kwargs)
                if extentity and extentity.etype == etype:
                    extentity.values.setdefault(rtype, set()).add(self.record.extid)
                return extentity

        return wraps(func)(wrapper)

    return decorator


def add_child_for(etype, relation, builder):
    """Handle import of child generated by `builder` function for `etype` ExtEntity
    that is yielded by decorated method.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, elem):
            build_child = getattr(self, builder)
            for extentity in func(self, elem):
                if extentity.etype == etype:
                    for child in build_child(elem):
                        extentity.values.setdefault(relation, set()).add(child.extid)
                        yield child
                yield extentity

        return wrapper

    return decorator


add_citations_for = partial(
    add_child_for, relation="has_citation", builder="build_citation"
)
add_events_for = partial(
    add_child_for, relation="has_event", builder="parse_history_chronitems"
)
add_names_for = partial(
    add_child_for, relation="simple_name_relation", builder="build_name_child"
)
add_dates_for = partial(
    add_child_for, relation="date_relation", builder="build_date_entry"
)
add_place_entries_for = partial(
    add_child_for, relation="place_entry_relation", builder="build_place_entry"
)


def require_tag(tagname):
    """Method decorator handling a mandatory tag within a XML element."""

    def warn(self, elem):
        self.import_log.record_warning(
            self._("expecting a %s tag in element %s, found none")
            % (tagname, elem.tag),
            line=elem.sourceline,
        )

    def decorator(func):
        # pylint: disable=protected-access
        if inspect.isgeneratorfunction(func):

            def wrapped(self, elem, *args, **kwargs):
                if self._elem_find(elem, tagname) is None:
                    warn(self, elem)
                    return
                for extentity in func(self, elem, *args, **kwargs):
                    yield extentity

        else:

            def wrapped(self, elem, *args, **kwargs):
                if self._elem_find(elem, tagname) is None:
                    warn(self, elem)
                    return None
                return func(self, elem, *args, **kwargs)

        return wraps(func)(wrapped)

    return decorator


def trace_extentity(instance):
    """Decorator for `build_` methods tracing ExtEntities built from a given
    XML element.
    """

    def decorator(func):
        if inspect.isgeneratorfunction(func):

            def wrapper(elem, *args, **kwargs):
                for extentity in func(elem, *args, **kwargs):
                    instance.record_visited(elem, extentity)
                    yield extentity

        else:

            def wrapper(elem, *args, **kwargs):
                extentity = func(elem, *args, **kwargs)
                if extentity is not None:
                    instance.record_visited(elem, extentity)
                return extentity

        return wraps(func)(wrapper)

    return decorator


def equivalent_concept(tagname, etype):
    """Method decorator indicating that a sub-tag may have a vocabularySource attribute indicating
    that a relation to some ExternalUri object should be drown from any entity of type `etype` built
    by decorated method.
    """

    def decorator(func):
        @wraps(func)
        def wrapped(self, elem, *args, **kwargs):
            subelem = self._elem_find(elem, tagname)
            if subelem is not None:
                extid = subelem.attrib.get("vocabularySource")
                if extid:
                    yield external_uri(extid)
            else:
                extid = None

            def update_extentity(extentity):
                if extid is not None and extentity.etype == etype:
                    extentity.values["equivalent_concept"] = set([extid])

            if inspect.isgeneratorfunction(func):
                for extentity in func(self, elem, *args, **kwargs):
                    update_extentity(extentity)
                    yield extentity
            else:
                extentity = func(self, elem, *args, **kwargs)
                update_extentity(extentity)
                yield extentity

        return wrapped

    return decorator


class EACCPFImporter(object):
    """Importer for EAC-CPF data.

    The importer will generate `extid`s using the `extid_generator` function
    if specified or use `uuid.uuid4` to generate unique `extid`s.

    During import the `record` attribute is set to the external entity of the
    imported AuthorityRecord.

    Ref: http://eac.staatsbibliothek-berlin.de/fileadmin/user_upload/schema/cpfTagLibrary.html
    """

    def __init__(self, fpath, import_log, _=str, extid_generator=None):
        try:
            tree = etree.parse(fpath)
        except etree.XMLSyntaxError as exc:
            raise InvalidXML(str(exc))
        self._ = _
        self._root = tree.getroot()
        self.namespaces = self._root.nsmap.copy()
        # remove default namespaces, not supported by .xpath method we'll use later
        self.namespaces.pop(None, None)
        self.namespaces["eac"] = "urn:isbn:1-931666-33-4"
        self.namespaces.setdefault("xlink", "http://www.w3.org/1999/xlink")
        self.import_log = import_log
        if extid_generator is None:

            def extid_generator():
                return str(uuid4())

        self._gen_extid = extid_generator
        self.record = ExtEntity("AuthorityRecord", None, {})
        # Store a mapping of XML elements to produced ExtEntities
        self._visited = {}

    def __getattribute__(self, name):
        attr = super(EACCPFImporter, self).__getattribute__(name)
        if name.startswith("build_"):
            return trace_extentity(self)(attr)
        return attr

    def values_from_attrib(self, elem, name_attrib_tuples):
        values = {}
        for var_name, attrib_name in name_attrib_tuples:
            attrib_value = elem.attrib.get(attrib_name)
            if attrib_value:
                values[var_name] = set([str(attrib_value)])
        return values

    def values_from_xpaths(self, elem, name_path_tuples):
        values = {}
        for var_name, var_path in name_path_tuples:
            var = self._elem_find(elem, var_path)
            if var is not None and var.text:
                values[var_name] = set([str(var.text)])
        return values

    def parse_items(self, elem):
        values = []
        new_elem = copy.deepcopy(elem)
        for list_elem in self._elem_findall(new_elem, "eac:list"):
            list_elem.tag = "ul"
            for child in self._elem_findall(list_elem, "eac:item"):
                child.tag = "li"
            values.append(list_elem)
        result = "\n".join(etree.tostring(item, encoding="unicode") for item in values)
        filtered = [
            e for e in re.split(r"xmlns\S*\"", result) if not e.startswith("xmlns")
        ]
        return "".join(filtered)

    def record_visited(self, elem, extentity):
        assert extentity.extid, extentity
        self._visited.setdefault(elem, set([])).add(extentity.extid)

    def not_visited(self):
        """Yield (tagname, sourceline) items corresponding to XML elements not
        used to build any ExtEntity.
        """
        visited = self._visited
        ns = self.namespaces["eac"]
        queue = deque(self._root)
        # These elements contain other ones which are known to be handled.
        container_tags = [
            "control",
            "cpfDescription",
            "identity",
            "maintenanceHistory",
            "sources",
            "description",
            "mandates",
            "places",
            "legalStatuses",
            "occupations",
            "relations",
        ]
        containers = ["{{{0}}}{1}".format(ns, tag) for tag in container_tags]
        while queue:
            elem = queue.popleft()
            if not isinstance(elem, etree._Element) or isinstance(elem, etree._Comment):
                continue
            if elem in visited:
                continue
            if elem.tag not in containers:
                yield elem.tag.replace("{" + ns + "}", ""), elem.sourceline
            else:
                queue.extend(elem)

    def _elem_find(self, elem, path, method="find"):
        """Wrapper around lxml.etree.Element find* methods to support
        namespaces also for old lxml versions.
        """
        finder = getattr(elem, method)
        try:
            return finder(path, self.namespaces)
        except TypeError:
            # In old lxml, find() does not accept namespaces argument.
            path = path.split(":", 1)
            try:
                ns, path = path
            except ValueError:
                # path has no namespace
                pass
            else:
                path = "{" + self.namespaces[ns] + "}" + path
            return finder(path)

    def _elem_findall(self, *args):
        return self._elem_find(*args, method="findall")

    @filter_empty
    def external_entities(self):
        """Parse a EAC XML file to and yield external entities."""
        # control element.
        control = self._elem_find(self._root, "eac:control")
        if control is None:
            raise MissingTag("control")
        for extentity in self.parse_control(control):
            yield extentity
        # Records (identity tags) are within cpfDescription tag.
        cpf_desc = self._elem_find(self._root, "eac:cpfDescription")
        if cpf_desc is None:
            raise MissingTag("cpfDescription")
        # identity element.
        identity = self._elem_find(cpf_desc, "eac:identity")
        if identity is None:
            raise MissingTag("identity", "cpfDescription")
        for extentity in self.parse_identity(identity):
            yield extentity
        # description element.
        description = self._elem_find(cpf_desc, "eac:description")
        if description is not None:
            for extentity in self.parse_description(description):
                yield extentity
        # relations element.
        for extentity in self.parse_relations(cpf_desc):
            yield extentity
        # Record is complete.
        self.record_visited(self._root, self.record)
        yield self.record

    def parse_identity(self, identity):
        """Parse the `identity` tag and yield external entities, possibly
        updating record's `values` dict.
        """
        # entityId
        isni = self._elem_find(identity, "eac:entityId")
        if isni is not None and isni.text:
            self.record_visited(isni, self.record)
            self.record.values["isni"] = set([str(isni.text)])
        # entityType
        akind = self._elem_find(identity, "eac:entityType")
        if akind is None:
            raise MissingTag("entityType", "identity")
        agent_kind = self.build_agent_kind(akind)
        yield agent_kind
        self.record.values["agent_kind"] = set([agent_kind.extid])
        name_entry = None
        name_entries = self._elem_findall(identity, "eac:nameEntry")
        if not name_entries:
            raise MissingTag("nameEntry", "identity")
        for name_entry in name_entries:
            for extentity in self.build_name_entry(name_entry):
                yield extentity
        parallel_name_entries = self._elem_findall(identity, "eac:nameEntryParallel")
        for parallel_name_entry in parallel_name_entries:
            for extentity in self.build_parallel(parallel_name_entry):
                yield extentity

    @relate_to_record_through("ParallelNames", "parallel_names_of")
    @add_names_for("ParallelNames")
    @add_dates_for("ParallelNames")
    def build_parallel(self, elem):
        """For each nameEntryParallel build a new object linked to the
        EAC-CPF document and get relations for all childrens"""
        values = self.values_from_xpaths(
            elem,
            (
                ("authorized_form", "eac:authorizedForm"),
                ("alternative_form", "eac:alternativeForm"),
            ),
        )
        yield ExtEntity("ParallelNames", self._gen_extid(), values)

    @filter_empty
    def extract_dates_from(self, elem, tag):
        for date in self.find_nested(elem, "eac:date", tag):
            yield ExtEntity(
                "DateEntity",
                self._gen_extid(),
                {
                    "start_date": set([self.parse_date(date)]),
                    "end_date": set([self.parse_date(date)]),
                },
            )
        for date_range in self.find_nested(elem, "eac:dateRange", tag):
            yield ExtEntity(
                "DateEntity", self._gen_extid(), self.parse_daterange(date_range)
            )

    @filter_empty
    @elem_maybe_none
    def build_date_entry(self, elem):
        """Build DateEntitys linked to a parent entity"""
        for usedates in self._elem_findall(elem, "eac:useDates"):
            for extentity in self.build_date_entry(usedates):
                yield extentity
        if self._elem_find(elem, "eac:dateSet"):
            for date_set in self._elem_findall(elem, "eac:dateSet"):
                for extentity in self.extract_dates_from(date_set, "eac:dateSet"):
                    yield extentity
        for entity in self.extract_dates_from(elem, "eac:test"):
            yield entity

    @filter_empty
    @filter_none
    @elem_maybe_none
    def build_name_child(self, elem):
        """Build NameEntry external entity"""
        for elem in self._elem_findall(elem, "eac:nameEntry"):
            values = self.values_from_attrib(
                elem, (("language", "lang"), ("script_code", "scriptCode"))
            )
            values.update(
                self.values_from_xpaths(
                    elem,
                    (
                        ("preferred_form", "eac:preferredForm"),
                        ("alternative_form", "eac:alternativeForm"),
                        ("authorized_form", "eac:authorizedForm"),
                    ),
                )
            )
            parts = self._elem_findall(elem, "eac:part")
            if not parts:
                raise MissingTag("part", "nameEntry")
            values["parts"] = {", ".join(str(p.text) for p in parts)}
            yield ExtEntity("NameEntry", self._gen_extid(), values)

    def parse_languages(self, elems):
        """Add a `languages` on AuthorityRecord"""
        languages = []
        for elem in elems:
            language = self._elem_find(elem, "eac:language")
            if language is not None and language.text.strip():
                languages.append(language.text.strip())
        if languages:
            self.record.values["languages"] = set([", ".join(languages)])

    @filter_none
    def parse_description(self, description):
        """Parse the `description` tag and yield external entities, possibly
        updating record's `values` dict.
        """
        # dates.
        daterange = description.xpath(
            "eac:existDates/eac:dateRange", namespaces=self.namespaces
        )
        if daterange:
            elem = daterange[0]
            self.record_visited(elem, self.record)
            self.record_visited(elem.getparent(), self.record)
            dates = self.parse_daterange(elem)
            if dates:
                self.record.values.update(dates)
        # address.
        for place in self.find_nested(description, "eac:place", "eac:places"):
            for extentity in self.build_place(place):
                yield extentity
        # additional EAC-CPF information.
        for legal_status in self.find_nested(
            description, "eac:legalStatus", "eac:legalStatuses"
        ):
            for extentity in self.build_legal_status(legal_status):
                yield extentity
        # mandate
        for mandate in self.find_nested(description, "eac:mandate", "eac:mandates"):
            for extentity in self.build_mandate(mandate):
                yield extentity
        # languagesUsed
        languages = self.find_nested(
            description, "eac:languageUsed", "eac:languagesUsed"
        )
        if languages:
            self.parse_languages(languages)
        # history
        for history in self._elem_findall(description, "eac:biogHist"):
            for extentity in self.build_history(history):
                yield extentity
        # structure
        for structure in self._elem_findall(description, "eac:structureOrGenealogy"):
            for extentity in self.build_structure(structure):
                yield extentity
        # function
        for function in self.find_nested(description, "eac:function", "eac:functions"):
            for extentity in self.build_function(function):
                yield extentity
        # occupation
        for occupation in self.find_nested(
            description, "eac:occupation", "eac:occupations"
        ):
            for extentity in self.build_occupation(occupation):
                yield extentity
        # general context
        for context in self._elem_findall(description, "eac:generalContext"):
            for extentity in self.build_generalcontext(context):
                yield extentity

    def find_nested(self, elem, tagname, innertag):
        """Return a list of element with `tagname` within `element` possibly
        nested within `innertag`.
        """
        all_elems = self._elem_findall(elem, tagname)
        wrapper = self._elem_find(elem, innertag)
        if wrapper is not None:
            all_elems.extend(self._elem_findall(wrapper, tagname))
        return all_elems

    def parse_tag_description(
        self, elem, tagname="eac:descriptiveNote", attrname="description"
    ):
        """Return a dict with `attrname` and `attrname`_format retrieved from
        a description-like tag.
        """
        elems = self._elem_findall(elem, tagname)
        if len(elems) > 1:
            self.import_log.record_warning(
                self._(
                    "found multiple %s tag within %s element, only one will be " "used."
                )
                % (tagname, elem.tag),
                line=elem.sourceline,
            )
        elem = elems[0] if elems else None
        values = {}
        if elem is not None:
            parsed = self.parse_tag_content(elem)
            values.update(
                zip((attrname, attrname + "_format"), (set([p]) for p in parsed))
            )
        return values

    def parse_tag_content(self, elem):
        """Parse the content of an element be it plain text or HTML and return
        the content along with MIME type.
        """
        assert elem is not None, "unexpected empty element"
        text = elem.text.strip() if elem.text else None
        if text:
            desc, desc_format = str(text), "text/plain"
        else:
            ptag = "{%(eac)s}p" % self.namespaces
            desc = "\n".join(
                etree.tostring(child, encoding=str, method="html").strip()
                for child in elem.iterchildren(ptag)
                if len(child) != 0 or child.text
            )
            if desc:
                desc_format = "text/html"
            else:
                self.import_log.record_warning(
                    self._(
                        "element %s has no text nor children, no content " "extracted"
                    )
                    % elem.tag,
                    line=elem.sourceline,
                )
                desc, desc_format = None, None
        return desc, desc_format

    @relate_to_record_through("NameEntry", "name_entry_for")
    @add_dates_for("NameEntry")
    def build_name_entry(self, element):
        """Build a NameEntry external entity."""
        self.record_visited(element, self.record)
        parts = self._elem_findall(element, "eac:part")
        if not parts:
            raise MissingTag("part", "nameEntry")
        # Join all "part" tags into a single "parts" attribute.
        values = {"parts": set([", ".join(str(p.text) for p in parts)])}
        # Consider first authorizedForm and then alternativeForm, missing
        # possible combinations which cannot be handled until the model is
        # complete.
        if self._elem_find(element, "eac:authorizedForm") is not None:
            values["form_variant"] = set(["authorized"])
        elif self._elem_find(element, "eac:alternativeForm") is not None:
            values["form_variant"] = set(["alternative"])
        yield ExtEntity("NameEntry", self._gen_extid(), values)

    @elem_maybe_none
    def build_agent_kind(self, elem):
        """Build a AgentKind external entity"""
        # Map EAC entity types to our terminolgy.
        kind = TYPE_MAPPING.get(elem.text, "unknown-agent-kind")
        if kind == "unknown-agent-kind":
            msg = self._("unexpected entity type {}").format(elem.text)
            self.import_log.record_warning(msg, line=elem.sourceline)
        agentkind_id = "agentkind/" + kind
        return ExtEntity("AgentKind", agentkind_id, {"name": set([str(kind)])})

    @elem_maybe_none
    @relate_to_record_through("LegalStatus", "legal_status_agent")
    @filter_empty
    @add_citations_for("LegalStatus")
    @add_place_entries_for("LegalStatus")
    @add_items_for("LegalStatus")
    @add_dates_for("LegalStatus")
    @equivalent_concept("eac:term", "LegalStatus")
    def build_legal_status(self, elem, **kwargs):
        """Build a `LegalStatus` external entity.

        Extra `kwargs` are passed to `parse_tag_description`.
        """
        values = self.parse_tag_description(elem, **kwargs)
        term = self._elem_find(elem, "eac:term")
        if term is not None and term.text:
            values["term"] = set([str(term.text)])
        yield ExtEntity("LegalStatus", self._gen_extid(), values)

    @elem_maybe_none
    @relate_to_record_through("Mandate", "mandate_agent")
    @filter_empty
    @add_citations_for("Mandate")
    @add_place_entries_for("Mandate")
    @add_items_for("Mandate")
    @add_dates_for("Mandate")
    @equivalent_concept("eac:term", "Mandate")
    def build_mandate(self, elem, **kwargs):
        """Build a `Mandate` external entity.

        Extra `kwargs` are passed to `parse_tag_description`.
        """
        values = self.parse_tag_description(elem, **kwargs)
        term = self._elem_find(elem, "eac:term")
        if term is not None and term.text:
            values["term"] = set([str(term.text)])
        yield ExtEntity("Mandate", self._gen_extid(), values)

    @elem_maybe_none
    def build_citation(self, elem):
        """Build a `Citation` external entity."""
        for citation_elem in self._elem_findall(elem, "eac:citation"):
            note = citation_elem.text.strip() if citation_elem.text else ""
            uri = citation_elem.attrib.get("{%(xlink)s}href" % self.namespaces)
            if not note and not uri:
                msg = self._("element {0} has no text nor (valid) link").format(
                    etree.tostring(citation_elem)
                )
                self.import_log.record_warning(msg, line=citation_elem.sourceline)
                return
            values = {}
            if uri:
                values["uri"] = set([str(uri)])
            if note:
                values["note"] = set([str(note)])
                if "<span>" in note:
                    values["note_format"] = set(["text/html"])
            yield ExtEntity("Citation", self._gen_extid(), values)

    @relate_to_record_through("History", "history_agent")
    @add_citations_for("History")
    @add_items_for("History")
    @elem_maybe_none
    def build_history(self, elem):
        """Build a `History` external entity."""
        abstract = self._elem_find(elem, "eac:abstract")
        desc, desc_format = self.parse_tag_content(elem)
        if desc:
            values = {"text": set([desc]), "text_format": set([desc_format])}
            if abstract is not None and abstract.text:
                values["abstract"] = set([str(abstract.text)])
            history = ExtEntity("History", self._gen_extid(), values)
            for child in self.parse_history_chronitems(elem):
                if child.etype == "HistoricalEvent":
                    history.values.setdefault("has_event", set()).add(child.extid)
                yield child
            yield history

    def parse_history_chronitems(self, elem):
        """Build en `Event` external entity."""
        for citem in self._elem_findall(elem, ".//eac:chronItem"):
            for extentity in self.build_event(citem):
                yield extentity

    @add_dates_for("HistoricalEvent")
    @add_place_entries_for("HistoricalEvent")
    @filter_none
    @filter_empty
    @elem_maybe_none
    def build_event(self, elem):
        """Build a `HistoricalEvent` external entity."""
        values = {}
        event = self._elem_find(elem, "eac:event")
        if event is not None and event.text:
            values["event"] = set([str(event.text)])
        yield ExtEntity("HistoricalEvent", self._gen_extid(), values)

    @elem_maybe_none
    @relate_to_record_through("Structure", "structure_agent")
    @add_citations_for("Structure")
    @add_items_for("Structure")
    def build_structure(self, elem):
        """Build a `Structure` external entity."""
        desc, desc_format = self.parse_tag_content(elem)
        if desc:
            values = {
                "description": set([desc]),
                "description_format": set([desc_format]),
            }
            yield ExtEntity("Structure", self._gen_extid(), values)

    @relate_to_record_through("AgentPlace", "place_agent")
    @filter_empty
    @add_citations_for("AgentPlace")
    @add_dates_for("AgentPlace")
    @add_items_for("AgentPlace")
    def build_place(self, elem):
        """Build a AgentPlace external entity"""
        values = {}
        role = self._elem_find(elem, "eac:placeRole")
        if role is not None:
            values["role"] = set([str(role.text)])
        for address in self._elem_findall(elem, "eac:address"):
            for extentity in self.build_address(address):
                if extentity.values:
                    values["place_address"] = set([extentity.extid])
                    yield extentity
        place = ExtEntity("AgentPlace", self._gen_extid(), values)
        for child in self.build_place_entry(elem):
            if child.etype == "PlaceEntry":
                place.values.setdefault("place_entry_relation", set()).add(child.extid)
            yield child
        yield place

    @filter_empty
    @equivalent_concept("eac:placeEntry", "PlaceEntry")
    def build_place_entry(self, elem):
        for entry in self._elem_findall(elem, "eac:placeEntry"):
            values = {"name": set([str(entry.text)])}
            values.update(
                self.values_from_attrib(
                    entry,
                    (
                        ("local_type", "localType"),
                        ("longitude", "longitude"),
                        ("latitude", "latitude"),
                    ),
                )
            )
            yield ExtEntity("PlaceEntry", self._gen_extid(), values)

    def build_address(self, elem):
        """Build `PostalAddress`s external entity"""
        address_entity = {}
        address_lines = []
        for line in self._elem_findall(elem, "eac:addressLine"):
            address_lines.append(str(line.text))
            if "localType" in line.attrib:
                attr = dict(ADDRESS_MAPPING).get(line.attrib["localType"])
                if attr:
                    address_entity.setdefault(attr, set()).add(str(line.text))
        if address_lines:
            address_entity["raw_address"] = set(["\n".join(address_lines)])
        yield ExtEntity("PostalAddress", self._gen_extid(), address_entity)

    @relate_to_record_through("AgentFunction", "function_agent")
    @filter_empty
    @add_citations_for("AgentFunction")
    @add_place_entries_for("AgentFunction")
    @add_dates_for("AgentFunction")
    @add_items_for("AgentFunction")
    @equivalent_concept("eac:term", "AgentFunction")
    def build_function(self, elem):
        """Build a `AgentFunction`s external entities"""
        values = self.parse_tag_description(elem)
        term = self._elem_find(elem, "eac:term")
        if term is not None:
            values["name"] = set([str(term.text)])
        yield ExtEntity("AgentFunction", self._gen_extid(), values)

    @relate_to_record_through("Occupation", "occupation_agent")
    @filter_empty
    @add_citations_for("Occupation")
    @add_place_entries_for("Occupation")
    @add_items_for("Occupation")
    @add_dates_for("Occupation")
    @equivalent_concept("eac:term", "Occupation")
    def build_occupation(self, elem):
        """Build a `Occupation`s external entities"""
        values = self.parse_tag_description(elem)
        term = self._elem_find(elem, "eac:term")
        if term is not None:
            values["term"] = set([str(term.text)])
        yield ExtEntity("Occupation", self._gen_extid(), values)

    @relate_to_record_through("GeneralContext", "general_context_of")
    @add_citations_for("GeneralContext")
    @add_items_for("GeneralContext")
    def build_generalcontext(self, elem):
        """Build a `GeneralContext` external entity"""
        content, content_format = self.parse_tag_content(elem)
        if content:
            values = {
                "content": set([content]),
                "content_format": set([content_format]),
            }
            yield ExtEntity("GeneralContext", self._gen_extid(), values)

    @elem_maybe_none
    def parse_daterange(self, elem):
        """Parse a `dateRange` tag and return a dict mapping `start_date` and
        `end_date` to parsed date range.
        """
        values = {}
        for eactag, attrname in zip(
            ("eac:fromDate", "eac:toDate"), ("start_date", "end_date")
        ):
            date = self.parse_date(self._elem_find(elem, eactag))
            if date:
                values[attrname] = set([date])
        return values

    @elem_maybe_none
    def parse_date(self, elem):
        """Parse a date-like element"""

        def record_warning(msg):
            self.import_log.record_warning(
                msg % {"e": etree.tostring(elem)}, line=elem.sourceline
            )

        standard_date = elem.attrib.get("standardDate")
        if standard_date:
            date = standard_date
        else:
            for attr in ("notBefore", "notAfter"):
                if elem.attrib.get(attr):
                    record_warning(
                        self._(
                            "found an unsupported %s attribute in date "
                            "element %%(e)s"
                        )
                        % attr
                    )
            # Using element's text.
            date = elem.text
            if not date:
                record_warning(self._("no date specified"))
                return None
        # Set a default value for month and day; the year should always be
        # given.
        default = datetime.datetime(9999, 1, 1)
        try:
            pdate = parse_date(date, default=default)
        except ValueError:
            record_warning(self._("could not parse date %(e)s"))
            return None
        except Exception as exc:  # pylint: disable=broad-except
            # Usually a bug in dateutil.parser.
            record_warning(
                self._("unexpected error during parsing of date %%(e)s: %s")
                % to_unicode(exc)
            )
            logger = logging.getLogger("cubes.eac")
            logger.exception(self._("unhandled exception while parsing date %r"), date)
            return None
        else:
            if pdate.year == default.year:
                record_warning(self._("could not parse a year from date element %(e)s"))
                return None
            return pdate.date()

    def parse_relations(self, cpf_description):
        """Parse the `relations` tag and yield external entities, possibly
        updating record's `values` dict.
        """
        relations = self._elem_find(cpf_description, "eac:relations")
        if relations is None:
            return
        builders = (
            ("eac:cpfRelation", self.build_relation),
            ("eac:resourceRelation", self.build_resource_relation),
            ("eac:functionRelation", self.build_function_relation),
        )
        for xpath, builder in builders:
            for elem in self._elem_findall(relations, xpath):
                for extentity in builder(elem):
                    yield extentity

    @add_xml_wrap_for(
        "AssociationRelation",
        "ChronologicalRelation",
        "HierarchicalRelation",
        "IdentityRelation",
        "FamilyRelation",
    )
    @add_dates_for("AssociationRelation")
    @add_dates_for("ChronologicalRelation")
    @add_dates_for("HierarchicalRelation")
    @add_dates_for("IdentityRelation")
    @add_dates_for("FamilyRelation")
    @add_place_entries_for("AssociationRelation")
    @add_place_entries_for("ChronologicalRelation")
    @add_place_entries_for("HierarchicalRelation")
    @add_place_entries_for("IdentityRelation")
    @add_place_entries_for("FamilyRelation")
    def build_relation(self, elem):
        """Build a relation between records external entity (with proper type)."""
        relationship = elem.attrib.get("cpfRelationType")
        if relationship is None:
            self.import_log.record_warning(
                self._(
                    "found no cpfRelationType attribute in element %s, defaulting "
                    "to associative"
                )
                % etree.tostring(elem),
                line=elem.sourceline,
            )
            relationship = "associative"
        try:
            # "other_role" (resp. "agent_role") role designates the object of the relation (resp.
            # the agent described in the EAC-CPF instance).
            # See: http://eac.staatsbibliothek-berlin.de/fileadmin/user_upload/schema/cpfTagLibrary.html#cpfRelationType # noqa pylint: disable=line-too-long
            # In case the EAC relation is not qualified, we assume the object is the "parent" (or
            # oldest) in the relation.
            etype, other_role, agent_role = {
                "hierarchical": (
                    "HierarchicalRelation",
                    "hierarchical_parent",
                    "hierarchical_child",
                ),
                "hierarchical-parent": (
                    "HierarchicalRelation",
                    "hierarchical_parent",
                    "hierarchical_child",
                ),
                "hierarchical-child": (
                    "HierarchicalRelation",
                    "hierarchical_child",
                    "hierarchical_parent",
                ),
                "temporal": (
                    "ChronologicalRelation",
                    "chronological_predecessor",
                    "chronological_successor",
                ),
                "temporal-earlier": (
                    "ChronologicalRelation",
                    "chronological_predecessor",
                    "chronological_successor",
                ),
                "temporal-later": (
                    "ChronologicalRelation",
                    "chronological_successor",
                    "chronological_predecessor",
                ),
                "associative": (
                    "AssociationRelation",
                    "association_to",
                    "association_from",
                ),
                "identity": ("IdentityRelation", "identity_to", "identity_from"),
                "family": ("FamilyRelation", "family_to", "family_from"),
            }[relationship]
        except KeyError:
            self.import_log.record_warning(
                self._("unsupported cpfRelationType %s in element %s, skipping")
                % (relationship, etree.tostring(elem)),
                line=elem.sourceline,
            )
            return
        obj_uri = elem.attrib.get("{%(xlink)s}href" % self.namespaces)
        if not obj_uri:
            self.import_log.record_warning(
                self._(
                    "found a cpfRelation without any object (no "
                    "xlink:href attribute), skipping"
                ),
                line=elem.sourceline,
            )
            return
        yield external_uri(obj_uri)
        values = {agent_role: set([self.record.extid]), other_role: set([obj_uri])}
        rentry = self._elem_find(elem, "eac:relationEntry")
        if rentry is not None and rentry.text.strip():
            values["entry"] = set([str(rentry.text)])
        values.update(self.parse_tag_description(elem))
        yield ExtEntity(etype, self._gen_extid(), values)

    @add_place_entries_for("EACFunctionRelation")
    @add_dates_for("EACFunctionRelation")
    @add_xml_wrap_for("EACFunctionRelation")
    def build_function_relation(self, elem):
        """Build a relation between function entities

        yield an ExternalUri object, and an EACFunctionRelation
        object that make the link between the ExternalUri and
        the AuthorityRecord object"""
        values = self.parse_tag_description(elem)
        relationship = elem.attrib.pop("functionRelationType", None)
        obj_uri = elem.attrib.pop("{%(xlink)s}href" % self.namespaces, None)
        # Yield the ExternalUri object
        if obj_uri:
            yield external_uri(obj_uri)
            values.update({"function_relation_function": set([str(obj_uri)])})
        if relationship:
            values.update({"r_type": set([str(relationship)])})
        values.update(
            {
                "function_relation_agent": set([str(self.record.extid)]),
            }
        )
        values.update(
            self.values_from_xpaths(
                elem,
                (
                    ("place_entry", "eac:placeEntry"),
                    ("relation_entry", "eac:relationEntry"),
                ),
            )
        )
        if elem.attrib:
            attributes = json.dumps(dict(elem.attrib), sort_keys=True)
        else:
            attributes = json.dumps({})
        values.update({"xml_attributes": set([str(attributes)])})
        yield ExtEntity("EACFunctionRelation", self._gen_extid(), values)

    @add_place_entries_for("EACResourceRelation")
    @add_dates_for("EACResourceRelation")
    @add_xml_wrap_for("EACResourceRelation")
    def build_resource_relation(self, elem):
        """Build a `EACResourceRelation` external entity (along with
        ExternalUri entities).
        """
        obj_uri = elem.attrib.pop("{%(xlink)s}href" % self.namespaces, None)
        if obj_uri is None:
            self.import_log.record_warning(
                self._(
                    "found a resourceRelation without any object (no xlink:href "
                    "attribute), skipping"
                ),
                line=elem.sourceline,
            )
            return
        yield external_uri(obj_uri)
        values = {
            "resource_relation_resource": set([obj_uri]),
            "resource_relation_agent": set([self.record.extid]),
        }
        relation_entry = self._elem_find(elem, "eac:relationEntry")
        if relation_entry is not None:
            values["relation_entry"] = set([str(relation_entry.text)])
        resource_role = elem.attrib.pop("{%(xlink)s}role" % self.namespaces, None)
        if resource_role:
            values["resource_role"] = set([str(resource_role)])
        agent_role = elem.attrib.pop("resourceRelationType", None)
        if agent_role:
            values["agent_role"] = set([str(agent_role)])
        if elem.attrib:
            attributes = json.dumps(dict(elem.attrib), sort_keys=True)
        else:
            attributes = json.dumps({})
        values.update({"xml_attributes": set([str(attributes)])})
        values.update(self.parse_tag_description(elem))
        yield ExtEntity("EACResourceRelation", self._gen_extid(), values)

    @filter_none
    def parse_control(self, control):
        """Parse the `control` tag."""
        record_id = self._elem_find(control, "eac:recordId")
        if record_id is not None and record_id.text and record_id.text.strip():
            record_id = record_id.text.strip()
            self.record.extid = "authorityrecord-{}".format(record_id)
            self.record.values["record_id"] = set([to_unicode(record_id)])
            self.record_visited(record_id, self.record)
        else:
            raise InvalidEAC("recordId element in control tag is mandatory")
        for other_record_id in self._elem_findall(control, "eac:otherRecordId"):
            other_id = other_record_id.text.strip()
            if other_id:
                values = {
                    "eac_other_record_id_of": set([self.record.extid]),
                    "value": set([str(other_id)]),
                }
                if other_record_id.attrib.get("localType"):
                    values["local_type"] = set(
                        [str(other_record_id.attrib["localType"])]
                    )
                extentity = ExtEntity("EACOtherRecordId", self._gen_extid(), values)
                self.record_visited(other_record_id, extentity)
                yield extentity
        builders = (
            ("eac:sources/eac:source", self.build_source),
            (
                "eac:maintenanceHistory/eac:maintenanceEvent",
                self.build_maintenance_event,
            ),
            ("eac:conventionDeclaration", self.build_convention),
        )
        for xpath_str, builder in builders:
            for elem in control.xpath(xpath_str, namespaces=self.namespaces):
                for extentity in builder(elem):
                    yield extentity

    def build_maintenance_event(self, elem):
        """Parse a `maintenanceEvent` tag, yielding a prov:Activity external
        entity along with necessary Records.
        """
        values = {"generated": set([self.record.extid])}
        event_type = self.parse_event_type(self._elem_find(elem, "eac:eventType"))
        if event_type is not None:
            values["type"] = set([event_type])
        date = self._elem_find(elem, "eac:eventDateTime")
        if date is not None:
            dtattr = date.attrib.get("standardDateTime")
            if dtattr:
                try:
                    event_date = parse_date(dtattr)
                except ValueError:
                    self.import_log.record_warning(
                        self._("could not parse date from %s") % etree.tostring(date),
                        line=date.sourceline,
                    )
                else:
                    values["start"] = set([event_date])
                    values["end"] = set([event_date])
        values.update(self.parse_tag_description(elem, "eac:eventDescription"))
        values.update(self.values_from_xpaths(elem, (("agent", "eac:agent"),)))
        agent_type = self._elem_find(elem, "eac:agentType")
        values["agent_type"] = "unknown"
        if agent_type.text in {"human", "machine"}:
            values["agent_type"] = set([str(agent_type.text)])
        yield ExtEntity("Activity", self._gen_extid(), values)

    @relate_to_record_through("Convention", "convention_of")
    @add_citations_for("Convention")
    @filter_none
    @filter_empty
    @elem_maybe_none
    def build_convention(self, elem):
        """Build a `Convention` external entity"""
        values = self.parse_tag_description(elem)
        abbrev = self._elem_find(elem, "eac:abbreviation")
        if abbrev is not None and abbrev.text:
            values["abbrev"] = set([str(abbrev.text)])
        yield ExtEntity("Convention", self._gen_extid(), values)

    @elem_maybe_none
    def parse_event_type(self, elem):
        """Parse an `eventType` element and try to match a prov:type to build a
        prov:Activity.
        """
        event_type = elem.text.strip() if elem.text else None
        if event_type:
            type_mapping = MAINTENANCETYPE_MAPPING.copy()
            type_mapping["derived"] = "create"
            type_mapping["updated"] = "modify"
            try:
                event_type = type_mapping[event_type.lower()]
            except KeyError:
                self.import_log.record_warning(
                    self._(
                        "eventType %s does not match the PROV-O vocabulary, "
                        "respective Activity will not have a `type` attribute set."
                    )
                    % event_type,
                    line=elem.sourceline,
                )
                return None
            return event_type

    @relate_to_record_through("EACSource", "source_agent")
    @filter_empty
    @add_xml_wrap_for("EACSource")
    def build_source(self, elem):
        """Parse a `source` tag, yielding EACSource external entities."""
        values = self.parse_tag_description(elem)
        url = elem.attrib.get("{%(xlink)s}href" % self.namespaces)
        if url is not None:
            values["url"] = set([str(url)])
        entry = self._elem_find(elem, "eac:sourceEntry")
        if entry is not None and entry.text:
            values["title"] = set([str(entry.text)])
        yield ExtEntity("EACSource", self._gen_extid(), values)



# copyright 2015-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac unit tests for schema"""

import sqlite3
import unittest
from datetime import date
from contextlib import contextmanager

from cubicweb import ValidationError, Unauthorized
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_compound.utils import optional_relations, graph_relations
from cubicweb_eac import testutils, AuthorityRecordGraph


@contextmanager
def assertValidationError(self, cnx):
    with self.assertRaises(ValidationError) as cm:
        yield cm
        cnx.commit()
    cnx.rollback()


class SchemaConstraintsTC(CubicWebTC):
    assertValidationError = assertValidationError

    def test_postaladdress_no_required(self):
        """Make sure a PostalAddress can be "empty" (useful in data import
        context).
        """
        with self.admin_access.cnx() as cnx:
            ar = testutils.authority_record(cnx, "T-04", "test")
            cnx.create_entity(
                "AgentPlace",
                place_agent=ar,
                place_address=cnx.create_entity("PostalAddress"),
            )
            cnx.commit()
            # No assert, just make sure db integrity checks pass.

    def test_on_create_set_end_date_before_start_date(self):
        """create an entity whose end_date is before start_date.
        ValidationError expected
        """
        with self.admin_access.cnx() as cnx:
            with self.assertValidationError(cnx) as cm:
                testutils.authority_record(
                    cnx,
                    "T-04",
                    "Arthur",
                    start_date=date(524, 2, 9),
                    end_date=date(500, 7, 12),
                )
            self.assertIn("must be less than", str(cm.exception))

    def test_on_update_set_end_date_before_start_date(self):
        """create a valid entity and update it with a new end_date set before the start_date.
        ValidationError expected
        """
        if sqlite3.sqlite_version_info < (3, 7, 12):
            # with sqlite earlier than 3.7.12, boundary constraints are not checked by the database,
            # hence the constraint is only triggered on start_date modification
            self.skipTest("unsupported sqlite version")
        with self.admin_access.cnx() as cnx:
            agent = testutils.authority_record(
                cnx,
                "T-04",
                "Arthur",
                start_date=date(454, 2, 9),
                end_date=date(475, 4, 12),
            )
            cnx.commit()
            with self.assertValidationError(cnx) as cm:
                agent.cw_set(end_date=date(442, 7, 12))
            self.assertIn("must be less than", str(cm.exception))

    def test_on_update_set_start_date_after_end_date(self):
        """create an entity without start_date :
        No constraint on the end_date
        update the entity with a start_date set after the start_date :
        ValidationError expected
        """
        with self.admin_access.cnx() as cnx:
            agent = testutils.authority_record(
                cnx, "T-04", "Arthur", end_date=date(476, 2, 9)
            )
            cnx.commit()
            with self.assertValidationError(cnx) as cm:
                agent.cw_set(start_date=date(527, 4, 12))
            self.assertIn("must be less than", str(cm.exception))


class AuthorityRecordGraphTC(CubicWebTC):
    def test_graph_structure(self):
        graph = AuthorityRecordGraph(self.schema)
        expected = {
            "AgentFunction": {("function_agent", "subject"): set(["AuthorityRecord"])},
            "AgentPlace": {("place_agent", "subject"): set(["AuthorityRecord"])},
            "Citation": {
                ("has_citation", "object"): set(
                    [
                        "GeneralContext",
                        "Mandate",
                        "Occupation",
                        "AgentFunction",
                        "AgentPlace",
                        "History",
                        "LegalStatus",
                        "Convention",
                        "Structure",
                    ]
                )
            },
            "DateEntity": {
                ("date_relation", "object"): set(
                    [
                        "HistoricalEvent",
                        "AgentFunction",
                        "EACFunctionRelation",
                        "LegalStatus",
                        "Mandate",
                        "NameEntry",
                        "Occupation",
                        "AgentPlace",
                        "EACResourceRelation",
                        "ParallelNames",
                    ]
                )
            },
            "PlaceEntry": {
                ("place_entry_relation", "object"): set(
                    [
                        "HistoricalEvent",
                        "AgentFunction",
                        "EACFunctionRelation",
                        "LegalStatus",
                        "Mandate",
                        "Occupation",
                        "AgentPlace",
                        "EACResourceRelation",
                    ]
                )
            },
            "EACFunctionRelation": {
                ("function_relation_agent", "subject"): set(["AuthorityRecord"])
            },
            "EACOtherRecordId": {
                ("eac_other_record_id_of", "subject"): set(["AuthorityRecord"])
            },
            "EACResourceRelation": {
                ("resource_relation_agent", "subject"): set(["AuthorityRecord"])
            },
            "Convention": {("convention_of", "subject"): set(["AuthorityRecord"])},
            "EACSource": {("source_agent", "subject"): set(["AuthorityRecord"])},
            "HistoricalEvent": {("has_event", "object"): set(["History"])},
            "GeneralContext": {
                ("general_context_of", "subject"): set(["AuthorityRecord"])
            },
            "History": {("history_agent", "subject"): set(["AuthorityRecord"])},
            "LegalStatus": {
                ("legal_status_agent", "subject"): set(["AuthorityRecord"])
            },
            "Mandate": {("mandate_agent", "subject"): set(["AuthorityRecord"])},
            "NameEntry": {
                ("name_entry_for", "subject"): set(["AuthorityRecord"]),
                ("simple_name_relation", "object"): set(["ParallelNames"]),
            },
            "Occupation": {("occupation_agent", "subject"): set(["AuthorityRecord"])},
            "ParallelNames": {
                ("parallel_names_of", "subject"): set(["AuthorityRecord"])
            },
            "PostalAddress": {("place_address", "object"): set(["AgentPlace"])},
            "Structure": {("structure_agent", "subject"): set(["AuthorityRecord"])},
        }
        struct = dict(
            (k, dict((rel, set(targets)) for rel, targets in v.items()))
            for k, v in graph.parent_structure("AuthorityRecord").items()
        )
        self.assertEqual(struct, expected)

    def test_optional_relations(self):
        graph = AuthorityRecordGraph(self.schema)
        structure = graph.parent_structure("AuthorityRecord")
        opts = optional_relations(self.schema, structure)
        expected = {
            "NameEntry": set(
                [("name_entry_for", "subject"), ("simple_name_relation", "object")]
            )
        }
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = AuthorityRecordGraph(self.schema)
        structure = graph.parent_structure("AuthorityRecord")
        structurals, optionals, mandatories = graph_relations(self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)


class SecurityTC(CubicWebTC):
    """Test case for permissions set in the schema"""

    @contextmanager
    def assertUnauthorized(self, cnx):
        with self.assertRaises(Unauthorized) as cm:
            yield cm
            cnx.commit()
        cnx.rollback()

    def test_agentkind_type(self):
        with self.admin_access.cnx() as cnx:
            kind = cnx.find("AgentKind", name="person").one()
            # no one can update nor delete a kind
            with self.assertUnauthorized(cnx):
                kind.cw_set(name="gloups")
            with self.assertUnauthorized(cnx):
                kind.cw_delete()

        with self.admin_access.cnx() as cnx:
            with self.assertUnauthorized(cnx):
                cnx.create_entity("AgentKind", name="new")

    def test_agent_kind_relation(self):
        """Test we can only change kind from unknown to another."""
        with self.admin_access.cnx() as cnx:
            record = testutils.authority_record(
                cnx, "T-02", "bob", kind="unknown-agent-kind"
            )
            cnx.commit()
            record.cw_set(agent_kind=cnx.find("AgentKind", name="person").one())
            cnx.commit()
            with self.assertRaises(Unauthorized):
                record.cw_set(agent_kind=cnx.find("AgentKind", name="authority").one())

    def test_authority_record_base(self):
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, login="toto", groups=("users", "guests"))
            function = cnx.create_entity("AgentFunction", name="director")
            testutils.authority_record(
                cnx, "T-00", "admin record", reverse_function_agent=function
            )
            cnx.commit()

        with self.new_access("toto").cnx() as cnx:
            # user can create and modify its own records
            function = cnx.create_entity("AgentFunction", name="grouillot")
            record = testutils.authority_record(
                cnx, "T-02", "bob", reverse_function_agent=function
            )
            cnx.commit()
            function.cw_set(name="grouyo")
            cnx.commit()
            cnx.create_entity(
                "GeneralContext", content="plop", general_context_of=record
            )
            cnx.commit()
            # but not modify other's
            record = cnx.find("AuthorityRecord", has_text="admin record").one()
            with self.assertUnauthorized(cnx):
                record.cw_set(record_id="bobby")
            with self.assertUnauthorized(cnx):
                record.reverse_function_agent[0].cw_set(name="grouillot")
            with self.assertUnauthorized(cnx):
                function = cnx.create_entity("AgentFunction", name="grouillot")
                record.cw_set(reverse_function_agent=function)


if __name__ == "__main__":
    unittest.main()



# copyright 2015-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac views, mostly for import/export of AuthorityRecord from/to EAC.

Using only this cube, UI will be hardly usable to handle the complexity of the EAC model. You'll
find an UI implementation for EAC as part of the `saem_ref`_ cube.

.. _`saem_ref`: https://www.cubicweb.org/project/cubicweb-saem_ref
"""

from functools import partial
import os.path

from cubicweb import tags, _
from cubicweb.view import View
from cubicweb.predicates import (
    is_instance,
    match_user_groups,
    one_line_rset,
    score_entity,
)
from cubicweb.dataimport.importer import HTMLImportLog
from cubicweb.web import (
    action,
    component,
    formfields as ff,
    formwidgets as fw,
    httpcache,
)
from cubicweb.web.views import actions, calendar, cwsources, forms, idownloadable, uicfg

from cubicweb_skos import to_unicode

from cubicweb_eac import dataimport


abaa = uicfg.actionbox_appearsin_addmenu
pvs = uicfg.primaryview_section
pvdc = uicfg.primaryview_display_ctrl
afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs


# Edit / View setup

actions.CopyAction.__select__ &= ~is_instance("AuthorityRecord")

pvdc.tag_subject_of(("AuthorityRecord", "agent_kind", "*"), {"vid": "text"})

for etype, attr in (
    ("AuthorityRecord", "record_id"),
    ("AuthorityRecord", "isni"),
    ("AgentPlace", "name"),
    ("AgentPlace", "role"),
    ("Activity", "type"),
    ("Mandate", "term"),
    ("LegalStatus", "term"),
    ("EACResourceRelation", "agent_role"),
    ("EACResourceRelation", "resource_role"),
    ("EACSource", "title"),
    ("EACSource", "url"),
):
    affk.set_field_kwargs(etype, attr, widget=fw.TextInput({"size": 80}))


affk.set_field_kwargs("NameEntry", "form_variant", value="authorized")


for etype, attrs in [
    ("AgentFunction", ("name", "description")),
    ("AgentPlace", ("name", "role")),
    ("Mandate", ("term", "description")),
    ("LegalStatus", ("term", "description")),
    ("Occupation", ("term", "description")),
]:
    affk.set_fields_order(etype, ("vocabulary_source",) + attrs)
    if attrs[-1] == "description":
        affk.tag_attribute(
            (etype, "description"),
            {
                "help": _(
                    "let this unspecified to see the definition of the "
                    "related concept if a vocabulary is specifed"
                )
            },
        )

afs.tag_object_of(("*", "vocabulary_source", "*"), "main", "hidden")
afs.tag_subject_of(("*", "vocabulary_source", "*"), "main", "attributes")
afs.tag_object_of(("*", "equivalent_concept", "*"), "main", "hidden")

pvs.tag_subject_of(("AuthorityRecord", "postal_address", "*"), "hidden")

# Hide relation to the record for these entity types edition form.
for etype, rtype in (
    ("AgentPlace", "place_agent"),
    ("AgentFunction", "function_agent"),
    ("LegalStatus", "legal_status_agent"),
    ("Mandate", "mandate_agent"),
    ("History", "history_agent"),
    ("Structure", "structure_agent"),
    ("Occupation", "occupation_agent"),
    ("GeneralContext", "general_context_of"),
    ("EACSource", "source_agent"),
    ("EACResourceRelation", "resource_relation_agent"),
):
    afs.tag_subject_of((etype, rtype, "*"), "main", "hidden")

# TextArea for all description attributes where <p> is allowed.
for etype in (
    "AgentFunction",
    "Mandate",
    "Occupation",
    "Structure",
):
    affk.set_field_kwargs(etype, "description", widget=fw.TextArea())

# Citation
pvdc.tag_attribute(("Citation", "uri"), {"vid": "urlattr"})
# Generic rules for all entity types having an `has_citation` relationship.
afs.tag_subject_of(("*", "has_citation", "*"), "main", "inlined")
afs.tag_object_of(("*", "has_citation", "*"), "main", "hidden")
pvs.tag_object_of(("*", "has_citation", "Citation"), "hidden")

# AgentPlace
afs.tag_subject_of(("AgentPlace", "place_address", "PostalAddress"), "main", "inlined")
afs.tag_object_of(("AgentPlace", "place_address", "PostalAddress"), "main", "hidden")

# AgentFunction
affk.set_field_kwargs("AgentFunction", "name", widget=fw.TextInput({"size": 80}))


afs.tag_object_of(("*", "name_entry_for", "AuthorityRecord"), "main", "inlined")
afs.tag_subject_of(("NameEntry", "name_entry_for", "*"), "main", "hidden")

affk.set_fields_order("AuthorityRecord", [("name_entry_for", "object")])
pvdc.set_fields_order("AuthorityRecord", [("name_entry_for", "object")])
affk.set_field_kwargs("NameEntry", "parts", widget=fw.TextInput({"size": 80}))

affk.set_field_kwargs("EACOtherRecordId", "value", widget=fw.TextInput({"size": 80}))
affk.set_field_kwargs(
    "EACOtherRecordId", "local_type", widget=fw.TextInput({"size": 80})
)
afs.tag_subject_of(
    ("EACOtherRecordId", "eac_other_record_id_of", "*"), "main", "hidden"
)

pvdc.tag_attribute(("EACSource", "url"), {"vid": "urlattr"})


def unrelated_authorityrecord(rtype, form, field, **kwargs):
    """Choices function returning AuthorityRecord choices.

    Choices values are unrelated through `rtype` to edited entity thus
    filtering out possible ExternalURI targets. It also account for __linkto
    info to build a restricted choice when needed.
    """
    try:
        linkto = form.linked_to[rtype, "subject"][0]
    except (KeyError, IndexError):
        rset = form.edited_entity.unrelated(rtype, targettype="AuthorityRecord")
        return [(x.dc_title(), str(x.eid)) for x in rset.entities()]
    else:
        entity = form._cw.entity_from_eid(linkto)
        return [(entity.dc_title(), str(entity.eid))]


# Set fields order for authority records relations (autoform and primary view)
for etype, from_rdef, to_rdef in (
    ("AssociationRelation", "association_from", "association_to"),
    ("ChronologicalRelation", "chronological_predecessor", "chronological_successor"),
    ("HierarchicalRelation", "hierarchical_parent", "hierarchical_child"),
):
    affk.set_fields_order(
        etype, ("description", "start_date", "end_date", from_rdef, to_rdef)
    )
    for rtype in (from_rdef, to_rdef):
        affk.tag_subject_of(
            (etype, rtype, "*"), {"choices": partial(unrelated_authorityrecord, rtype)}
        )

    pvdc.set_fields_order(
        etype, ("description", "start_date", "end_date", from_rdef, to_rdef)
    )

# By default don't put anything in the 'add' submenu nor in the generic relation, expecting custom
# views to be implemented as in the saem_ref cube
abaa.tag_object_of(("*", "name_entry_for", "AuthorityRecord"), False)
for rtype in (
    "eac_other_record_id_of",
    "place_agent",
    "function_agent",
    "mandate_agent",
    "occupation_agent",
    "general_context_of",
    "legal_status_agent",
    "history_agent",
    "structure_agent",
    "source_agent",
    "resource_relation_agent",
    "association_from",
    "association_to",
    "chronological_predecessor",
    "chronological_successor",
    "hierarchical_parent",
    "hierarchical_child",
):
    abaa.tag_object_of(("*", rtype, "AuthorityRecord"), False)
    afs.tag_object_of(("*", rtype, "AuthorityRecord"), "main", "hidden")
abaa.tag_subject_of(("*", "has_citation", "*"), False)


class XMLWrapComponent(component.EntityCtxComponent):
    """CtxComponent to display xml_wrap of entities."""

    __select__ = component.EntityCtxComponent.__select__ & score_entity(
        lambda x: getattr(x, "xml_wrap", None)
    )
    __regid__ = "eac.xml_wrap"
    context = "navcontentbottom"
    title = _("xml_wrap")

    def render_body(self, w):
        sourcemt, targetmt = "text/xml", "text/html"
        data = self.entity.xml_wrap
        w(self.entity._cw_mtc_transform(data.getvalue(), sourcemt, targetmt, "utf-8"))


class AuthorityRecordICalendarableAdapter(calendar.ICalendarableAdapter):
    """ICalendarable adapter for AuthorityRecord entity type."""

    __select__ = calendar.ICalendarableAdapter.__select__ & is_instance(
        "AuthorityRecord"
    )

    @property
    def start(self):
        return self.entity.start_date

    @property
    def stop(self):
        return self.entity.end_date


# Import


class EACImportMixin(object):
    __regid__ = "eac.import"
    __select__ = match_user_groups("managers", "users")


class EACImportForm(EACImportMixin, forms.FieldsForm):
    """File import form for EAC-CPF"""

    filefield = ff.FileField(name="file", label=_("EAC-CPF file"), required=True)
    form_buttons = [fw.SubmitButton(label=_("do_import"))]

    @property
    def action(self):
        return self._cw.build_url(vid=self.__regid__)


class EACImportView(EACImportMixin, View):
    """EAC-CPF import controller"""

    def call(self):
        self.w(tags.h1(self._cw._("Importing an AuthorityRecord from a EAC-CPF file")))
        form = self._cw.vreg["forms"].select(self.__regid__, self._cw)
        if form.posting:
            posted = form.process_posted()
            stream = posted["file"]
            import_log = HTMLImportLog(os.path.basename(stream.filename))
            try:
                _, _, eid = self._cw.cnx.call_service(
                    self.__regid__,
                    stream=stream,
                    import_log=import_log,
                    **self.service_kwargs(posted)
                )
            except dataimport.InvalidXML as exc:
                msg = self._cw._("Invalid XML file")
                self.exception("error while importing %s", stream.filename)
                mtype = "danger"
                import_log.record_fatal(
                    self._cw._("xml syntax error: ") + to_unicode(exc)
                )
                self._cw.status_out = 400
            except dataimport.MissingTag as exc:
                if exc.tag_parent:
                    err = self._cw._(
                        "Missing tag %(tag)s within element %(parent)s in XML file"
                    )
                    params = {"tag": exc.tag, "parent": exc.tag_parent}
                else:
                    err = self._cw._("Missing tag %(tag)s in XML file")
                    params = {"tag": exc.tag}
                msg = err % params
                self.exception("error while importing %s", stream.filename)
                mtype = "danger"
                import_log.record_fatal(msg)
                self._cw.status_out = 400
            except Exception:  # pylint: disable=broad-except
                msg = self._cw._("EAC import failed")
                self.exception("error while importing %s", stream.filename)
                mtype = "danger"
                self._cw.status_out = 500
            else:
                record = self._cw.find("AuthorityRecord", eid=eid).one()
                msg = self._cw._("EAC-CPF import completed: %s") % record.view(
                    "oneline"
                )
                mtype = "success"
            self.w(tags.div(msg, klass="alert alert-%s" % mtype))
            if import_log.logs:
                self._cw.view(
                    "cw.log.table",
                    pyvalue=cwsources.log_to_table(self._cw, "".join(import_log.logs)),
                    default_level="Warning",
                    w=self.w,
                )
        else:
            form.render(w=self.w)

    def service_kwargs(self, posted):
        """Subclass access point to provide extra arguments to the service (e.g. saem_ref cube)."""
        return {}


# Export


class EACExportAction(action.Action):
    __regid__ = "eac.export"
    __select__ = (
        action.Action.__select__ & one_line_rset() & is_instance("AuthorityRecord")
    )

    title = _("EAC export")
    category = "moreactions"

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.__regid__)


class EACDownloadView(idownloadable.DownloadView):
    """EAC download view"""

    __regid__ = "eac.export"
    __select__ = one_line_rset() & is_instance("AuthorityRecord")

    http_cache_manager = httpcache.NoHTTPCacheManager
    adapter_id = "EAC-CPF"

    def set_request_content_type(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.adapter_id)
        self._cw.set_content_type(
            adapter.content_type,
            filename=adapter.file_name,
            encoding=adapter.encoding,
            disposition="attachment",
        )

    def call(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.adapter_id)
        self.w(adapter.dump())



for rtype in (
    "association_from",
    "association_to",
    "chronological_predecessor",
    "chronological_successor",
    "hierarchical_parent",
    "hierarchical_child",
):
    sync_schema_props_perms(rtype)



# copyright 2015-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac server objects"""

from cubicweb import ValidationError
from cubicweb.predicates import match_user_groups
from cubicweb.server import Service
from cubicweb.dataimport import RQLObjectStore
from cubicweb.dataimport.importer import ExtEntitiesImporter, cwuri2eid

from cubicweb_skos import to_unicode

from cubicweb_eac import dataimport


def init_extid2eid_index(cnx, source):
    """Return the dictionary mapping external id to eid for entities already in the database."""
    # only consider the system source for EAC related entity types
    extid2eid = cwuri2eid(cnx, dataimport.ETYPES_ORDER_HINT, source_eid=source.eid)
    # though concepts may come from any source
    extid2eid.update(cwuri2eid(cnx, ("Concept",)))
    return extid2eid


class EACImportService(Service):
    """Service to import an AuthorityRecord from an EAC XML file."""

    __regid__ = "eac.import"
    __select__ = match_user_groups("managers", "users")

    def call(self, stream, import_log, **kwargs):
        store = RQLObjectStore(self._cw)
        try:
            created, updated, record_eid, not_visited = self.import_eac_stream(
                stream, import_log, store, **kwargs
            )
        except Exception:
            self._cw.rollback()
            raise
        else:
            try:
                self._cw.commit()
            except ValidationError as exc:
                import_log.record_error("validation error: " + to_unicode(exc))
                raise
            else:
                import_log.record_info("%s entities created" % len(created))
                import_log.record_info("%s entities updated" % len(updated))
        msg = self._cw._("element {0} not parsed")
        for tagname, sourceline in not_visited:
            import_log.record_warning(msg.format(tagname, sourceline), line=sourceline)
        return created, updated, record_eid

    def external_entities_generator(self, stream, import_log):
        return dataimport.EACCPFImporter(stream, import_log, self._cw._)

    def import_eac_stream(self, stream, import_log, store, extid2eid=None, **kwargs):
        try:
            return self._import_eac_stream(
                stream, import_log, store, extid2eid=extid2eid, **kwargs
            )
        except Exception as exc:
            import_log.record_fatal(to_unicode(exc))
            raise

    def _import_eac_stream(self, stream, import_log, store, extid2eid=None, **kwargs):
        source = self._cw.repo.system_source
        if extid2eid is None:
            extid2eid = init_extid2eid_index(self._cw, source)
        importer = ExtEntitiesImporter(
            self._cw.vreg.schema,
            store,
            import_log=import_log,
            extid2eid=extid2eid,
            etypes_order_hint=dataimport.ETYPES_ORDER_HINT,
            **kwargs
        )
        generator = self.external_entities_generator(stream, import_log)
        extentities = self.external_entities_stream(
            generator.external_entities(), extid2eid
        )
        importer.import_entities(extentities)
        if generator.record is not None:
            extid = generator.record.extid
            record_eid = importer.extid2eid[extid]
        else:
            record_eid = None
        return importer.created, importer.updated, record_eid, generator.not_visited()

    def external_entities_stream(self, extentities, extid2eid):
        """Extracted method allowing to plug transformation into the external entities generator."""
        extentities = (
            ee
            for ee in extentities
            if not (ee.etype == "ExternalUri" and ee.extid in extid2eid)
        )

        def handle_agent_kind(extentities):
            """Create agent kind when necessary and remove them from the entity stream, allowing to
            set cwuri properly without attempt to update.
            """
            for extentity in extentities:
                if extentity.etype == "AgentKind":
                    assert (
                        extentity.extid in extid2eid
                    ), "unexpected agent kind {}".format(extentity)
                else:
                    yield extentity

        extentities = handle_agent_kind(extentities)
        return extentities



# pylint: disable=W0622
"""cubicweb-eac application packaging information"""

distname = "cubicweb-eac"
modname = "cubicweb_eac"  # required by apycot

numversion = (1, 5, 1)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "Implementation of Encoded Archival Context for CubicWeb"
web = "https://forge.extranet.logilab.fr/cubicweb/cubes/%s" % distname

__depends__ = {
    "cubicweb": ">= 3.26.0, < 3.38.0",
    "cubicweb-prov": ">= 0.4.0",
    "cubicweb-skos": ">= 1.3.0",
    "cubicweb-addressbook": ">=1.6.0",  # first release with python3 support
    "cubicweb-compound": ">= 0.6.0",
    "python-dateutil": None,
}
__recommends__ = {}

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: JavaScript",
]



# copyright 2016-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac schema"""

from yams.buildobjs import (
    EntityType,
    RelationDefinition,
    String,
    Date,
    Bytes,
    RichString,
    Float,
    ComputedRelation,
)
from yams.constraints import BoundaryConstraint, Attribute

from cubicweb import _
from cubicweb.schema import RRQLExpression, RQLVocabularyConstraint

from cubicweb_addressbook.schema import PostalAddress
from cubicweb_prov.schema import Activity


# Customization of addressbook schema.
for attrname in ("street", "city", "postalcode"):
    PostalAddress.get_relation(attrname).cardinality = "?1"

PostalAddress.add_relation(
    String(
        description=_("Attribute containing concatenated <addressLine> values"),
        indexed=True,
        fulltextindexed=True,
    ),
    name="raw_address",
)


def xml_wrap(cls):
    """Class decorator adding an `xml_wrap` attribute to an EntityType."""
    desc = _("XML elements not contained in EAC-CPF namespace")
    cls.add_relation(Bytes(description=desc), name="xml_wrap")
    return cls


def with_items(cls):
    """Class decorator adding an `items` attribute to an EntityType"""
    desc = _("HTML list of elements")
    cls.add_relation(RichString(description=desc), name="items")
    return cls


Activity.add_relation(
    String(
        description=_("the agent responsible for this activity"),
        indexed=True,
        fulltextindexed=True,
    ),
    name="agent",
)
Activity.add_relation(
    String(
        vocabulary=[_("human"), _("machine"), _("unknown")],
        description=_("the type of the agent linked to the activity"),
        indexed=True,
        fulltextindexed=True,
    ),
    name="agent_type",
)


class AuthorityRecord(EntityType):
    start_date = Date(
        constraints=[
            BoundaryConstraint(
                "<=",
                Attribute("end_date"),
                msg=_("start date must be less than end date"),
            )
        ]
    )
    end_date = Date()
    record_id = String(indexed=True, required=True, unique=True)
    isni = String(unique=True, description=_("International Standard Name Identifier"))
    languages = String()


class NameEntry(EntityType):
    """Represent a nameEntry tag of an EAC-CPF document."""

    language = String(
        fulltextindexed=True, description=_("language in wich the name is written")
    )
    preferred_form = String(
        fulltextindexed=True,
        description=_(
            "indicate which of the NameEntry in a ParallelNames"
            " is the preferred one "
        ),
    )
    alternative_form = String(
        fulltextindexed=True, description=_("indicate an alternative or variant forms")
    )
    authorized_form = String(
        fulltextindexed=True,
        description=_(
            "indicate the authorized form of the name "
            "according to a different set of rules"
        ),
    )
    script_code = String(
        fulltextindexed=True, description=_("language code of the language attribute")
    )
    parts = String(
        required=True,
        fulltextindexed=True,
        description=_("concatenation of part tags within a nameEntry"),
    )
    form_variant = String(
        internationalizable=True, vocabulary=[_("authorized"), _("alternative")]
    )


class ParallelNames(EntityType):
    authorized_form = String(fulltextindexed=True)
    alternative_form = String(fulltextindexed=True)


class simple_name_relation(RelationDefinition):
    subject = "ParallelNames"
    object = "NameEntry"
    cardinality = "*?"
    composite = "subject"
    fulltext_container = "subject"


class parallel_names_of(RelationDefinition):
    subject = "ParallelNames"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"


class name_entry_for(RelationDefinition):
    subject = "NameEntry"
    object = "AuthorityRecord"
    cardinality = "?+"
    composite = "object"
    fulltext_container = "object"
    inlined = True


class date_relation(RelationDefinition):
    subject = (
        "ParallelNames",
        "HierarchicalRelation",
        "HistoricalEvent",
        "ChronologicalRelation",
        "AssociationRelation",
        "AgentFunction",
        "EACFunctionRelation",
        "LegalStatus",
        "Mandate",
        "Occupation",
        "AgentPlace",
        "EACResourceRelation",
        "NameEntry",
        "IdentityRelation",
        "FamilyRelation",
    )
    object = "DateEntity"
    cardinality = "*1"
    composite = "subject"
    fulltext_container = "subject"


class PlaceEntry(EntityType):
    name = String(fulltextindexed=True)
    local_type = String(fulltextindexed=True)
    latitude = Float()
    longitude = Float()


class place_entry_relation(RelationDefinition):
    subject = (
        "HierarchicalRelation",
        "HistoricalEvent",
        "ChronologicalRelation",
        "AssociationRelation",
        "AgentFunction",
        "EACFunctionRelation",
        "LegalStatus",
        "Mandate",
        "Occupation",
        "AgentPlace",
        "EACResourceRelation",
        "IdentityRelation",
        "FamilyRelation",
    )
    object = "PlaceEntry"
    cardinality = "*1"
    composite = "subject"
    fulltext_container = "subject"


class EACOtherRecordId(EntityType):
    value = String(required=True, fulltextindexed=True, indexed=True)
    local_type = String(indexed=True)
    __unique_together__ = [("value", "eac_other_record_id_of")]


class eac_other_record_id_of(RelationDefinition):
    subject = "EACOtherRecordId"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True


@with_items
class AgentFunction(EntityType):
    """The function of an AuthorityRecord"""

    name = String(fulltextindexed=True, internationalizable=True)
    description = RichString(fulltextindexed=True)
    __unique_together__ = [("name", "function_agent")]


class function_agent(RelationDefinition):
    subject = "AgentFunction"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True


@with_items
class AgentPlace(EntityType):
    """Qualified relation between an AuthorityRecord and a PostalAddress"""

    role = String(
        description=_(
            "contextual role the address has in relation "
            'with the agent (e.g. "home")'
        ),
        internationalizable=True,
    )
    __unique_together__ = [("role", "place_agent", "place_address")]


class place_address(RelationDefinition):
    subject = "AgentPlace"
    object = "PostalAddress"
    cardinality = "?1"
    inlined = True
    composite = "subject"
    fulltext_container = "subject"


class place_agent(RelationDefinition):
    subject = "AgentPlace"
    object = "AuthorityRecord"
    cardinality = "1*"
    inlined = True
    composite = "object"
    fulltext_container = "object"


class postal_address(ComputedRelation):
    rule = "P place_agent S, P place_address O"


class AgentKind(EntityType):
    """Kind of an authority record (e.g. "person", "authority" or "family")"""

    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (),
        "update": (),
        "delete": (),
    }
    name = String(required=True, unique=True, internationalizable=True)


class agent_kind(RelationDefinition):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", "users"),
        "delete": (RRQLExpression('O name "unknown-agent-kind"'),),
    }
    subject = "AuthorityRecord"
    object = "AgentKind"
    cardinality = "1*"
    inlined = True


@with_items
class GeneralContext(EntityType):
    """Information about the general social and cultural context of an authority record"""

    content = RichString(fulltextindexed=True)


class general_context_of(RelationDefinition):
    subject = "GeneralContext"
    object = "AuthorityRecord"
    cardinality = "1*"
    inlined = True
    composite = "object"
    fulltext_container = "object"


class convention_of(RelationDefinition):
    subject = "Convention"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"


class _agent_relation(RelationDefinition):
    """Abstract relation between authority record"""

    subject = None
    object = ("AuthorityRecord", "ExternalUri")
    cardinality = "1*"
    inlined = True


@xml_wrap
class FamilyRelation(EntityType):
    """Family relation between authority records"""

    entry = String()
    description = RichString()


class family_from(_agent_relation):
    subject = "FamilyRelation"


class family_to(_agent_relation):
    subject = "FamilyRelation"


@xml_wrap
class IdentityRelation(EntityType):
    """Identity relation between authority records"""

    entry = String()
    description = RichString()


class identity_from(_agent_relation):
    subject = "IdentityRelation"


class identity_to(_agent_relation):
    subject = "IdentityRelation"


@xml_wrap
class AssociationRelation(EntityType):
    """Association relation between authority records"""

    entry = String()
    description = RichString()


class association_from(_agent_relation):
    subject = "AssociationRelation"


class association_to(_agent_relation):
    subject = "AssociationRelation"


@xml_wrap
class ChronologicalRelation(EntityType):
    """Chronological relation between authority records"""

    entry = String()
    description = RichString()


class chronological_predecessor(_agent_relation):
    subject = "ChronologicalRelation"


class chronological_successor(_agent_relation):
    subject = "ChronologicalRelation"


@xml_wrap
class HierarchicalRelation(EntityType):
    """Hierarchical relation between authority records"""

    entry = String()
    description = RichString()


class hierarchical_parent(_agent_relation):
    subject = "HierarchicalRelation"


class hierarchical_child(_agent_relation):
    subject = "HierarchicalRelation"


class generated(RelationDefinition):
    subject = "Activity"
    object = "AuthorityRecord"


class used(RelationDefinition):
    subject = "Activity"
    object = "AuthorityRecord"


@with_items
class Mandate(EntityType):
    """Reference text coming from an authority"""

    term = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


@with_items
class LegalStatus(EntityType):
    """Information relative to the legal status of an authority"""

    term = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


class Convention(EntityType):
    "Rules or conventions applied in creating the EAC-CPF instance"
    abbrev = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


@with_items
class History(EntityType):
    """Biographical or historical information"""

    abstract = RichString(fulltextindexed=True)
    text = RichString(fulltextindexed=True)


class HistoricalEvent(EntityType):
    """Events linked to an History object"""

    event = RichString(fulltextindexed=True)


class has_event(RelationDefinition):
    subject = "History"
    object = "HistoricalEvent"
    cardinality = "*1"
    composite = "subject"
    fulltext_container = "subject"
    description = _("HistoricalEvent with date for describing an historical event")


@with_items
class Structure(EntityType):
    """Information about the structure of an authority"""

    description = RichString(fulltextindexed=True)


@with_items
class Occupation(EntityType):
    term = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)


class occupation_agent(RelationDefinition):
    subject = "Occupation"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True
    description = _("occupation in which the person works or has worked")


class mandate_agent(RelationDefinition):
    subject = "Mandate"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True
    description = _("mandate of an authority record")


class legal_status_agent(RelationDefinition):
    subject = "LegalStatus"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True
    description = _("legal status of an authority record")


class structure_agent(RelationDefinition):
    subject = "Structure"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True
    description = _("information about the structure of an authority record")


class history_agent(RelationDefinition):
    subject = "History"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True
    description = _("information about the history of an authority record")


class Citation(EntityType):
    note = RichString()
    uri = String()


class has_citation(RelationDefinition):
    subject = (
        "GeneralContext",
        "Mandate",
        "Occupation",
        "History",
        "AgentFunction",
        "LegalStatus",
        "AgentPlace",
        "Convention",
        "Structure",
    )
    object = "Citation"
    cardinality = "*1"
    composite = "subject"
    fulltext_container = "subject"
    description = _("reference to an external citation resource")


@xml_wrap
class EACFunctionRelation(EntityType):
    """Represent a relation between an AuthorityRecord and a function"""

    r_type = String(
        internationalizable=True,
        description=_("type of relation the function has " "with the Authority"),
    )
    description = RichString(fulltextindexed=True)
    relation_entry = String(fulltextindexed=True)
    place_entry = String(fulltextindexed=True)
    xml_attributes = String(fulltextindexed=True)


class function_relation_agent(RelationDefinition):
    subject = "EACFunctionRelation"
    object = "AuthorityRecord"
    cardinality = "1*"
    inlined = True
    composite = "object"
    fulltext_container = "object"


class function_relation_function(RelationDefinition):
    subject = "EACFunctionRelation"
    object = ("AuthorityRecord", "ExternalUri")
    cardinality = "?*"
    inlined = True


@xml_wrap
class EACResourceRelation(EntityType):
    """Represent a relation between an AuthorityRecord and a remote resource in the
    EAC-CPF model.
    """

    relation_entry = RichString(
        fulltextindexed=True, description=_("name of the linked resource")
    )
    agent_role = String(
        description=_("type of relation the agent has to the resource"),
        internationalizable=True,
    )
    resource_role = String(
        description=_("type or nature of the remote resource"), internationalizable=True
    )
    description = RichString(
        fulltextindexed=True,
        description=_(
            "description of the relation between " "the agent and the resource"
        ),
    )
    xml_attributes = String(fulltextindexed=True)


class resource_relation_agent(RelationDefinition):
    subject = "EACResourceRelation"
    object = "AuthorityRecord"
    cardinality = "1*"
    inlined = True
    composite = "object"
    fulltext_container = "object"


class resource_relation_resource(RelationDefinition):
    subject = "EACResourceRelation"
    object = ("AuthorityRecord", "ExternalUri")
    cardinality = "1*"
    inlined = True


class DateEntity(EntityType):
    """Represent either a date tag or a dateRange tag
    for the representation of dateSets"""

    start_date = Date(
        constraints=[
            BoundaryConstraint(
                "<=",
                Attribute("end_date"),
                msg=_("start date must be less than end date"),
            )
        ]
    )
    end_date = Date()


@xml_wrap
class EACSource(EntityType):
    """A source used to establish the description of an AuthorityRecord"""

    title = String(fulltextindexed=True)
    url = String()
    description = RichString(fulltextindexed=True)


class source_agent(RelationDefinition):
    subject = "EACSource"
    object = "AuthorityRecord"
    cardinality = "1*"
    composite = "object"
    fulltext_container = "object"
    inlined = True


class vocabulary_source(RelationDefinition):
    subject = ("Mandate", "LegalStatus", "AgentFunction", "AgentPlace", "Occupation")
    object = "ConceptScheme"
    cardinality = "?*"


class equivalent_concept(RelationDefinition):
    subject = ("Mandate", "LegalStatus", "AgentFunction", "Occupation", "PlaceEntry")
    object = ("ExternalUri", "Concept")
    constraints = [RQLVocabularyConstraint("S vocabulary_source SC, O in_scheme SC")]
    cardinality = "?*"
    # relation with 'ExternalUri' as object can't be inlined because of a limitation of
    # data-import's (massive) store
    inlined = False


def post_build_callback(schema):
    from cubicweb_eac import AuthorityRecordGraph
    from cubicweb_compound.utils import (
        graph_set_etypes_update_permissions,
        graph_set_write_rdefs_permissions,
    )

    graph = AuthorityRecordGraph(schema)
    graph_set_etypes_update_permissions(schema, graph, "AuthorityRecord")
    graph_set_write_rdefs_permissions(schema, graph, "AuthorityRecord")



from cubicweb_compound import utils
from cubicweb_eac import AuthorityRecordGraph

add_cube("compound")

graph = AuthorityRecordGraph(schema)
structure = graph.parent_structure("AuthorityRecord")

optionals = utils.optional_relations(schema, structure)
for child in structure:
    sync_schema_props_perms(child, syncprops=False)

for rdef, parent_role in utils.mandatory_rdefs(schema, structure):
    sync_schema_props_perms((rdef.subject, rdef.rtype, rdef.object), syncprops=False)



# copyright 2015-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac test for views."""

import unittest

from cubicweb import Binary
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web.views.actions import CopyAction

from cubicweb_eac import testutils


class FuncViewsTC(CubicWebTC):
    def test_import_ok(self):
        regid = "eac.import"
        fname = "FRAD033_EAC_00001_simplified.xml"
        with self.admin_access.web_request() as req:
            # simply test the form properly render and is well formed
            self.view(regid, req=req, template=None)
            fields = {"file": (fname, open(self.datapath(fname), "rb"))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            req.view(regid)
            # several authority records are created (the main one + other for relations)
            self.assertTrue(req.find("AuthorityRecord"))

    def test_import_non_unique_isni(self):
        regid = "eac.import"
        fname = "FRAD033_EAC_00001_simplified.xml"
        with self.admin_access.client_cnx() as cnx:
            # ISNI is the same as the agent in EAC file.
            testutils.authority_record(cnx, "T-02", "bob", isni="22330001300016")
            cnx.commit()
        with self.admin_access.web_request() as req:
            # simply test the form properly render and is well formed
            self.view(regid, req=req, template=None)
            fields = {"file": (fname, open(self.datapath(fname), "rb"))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn("EAC import failed", html)
            # Still only one AuthorityRecord.
            rset = req.find("AuthorityRecord")
            self.assertEqual(len(rset), 1)

    def test_import_invalid_xml(self):
        regid = "eac.import"
        fname = "invalid_xml.xml"
        with self.admin_access.web_request() as req:
            fields = {"file": (fname, open(self.datapath(fname), "rb"))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn("Invalid XML file", html)

    def test_import_missing_tag(self):
        regid = "eac.import"
        fname = "missing_tag.xml"
        with self.admin_access.web_request() as req:
            fields = {"file": (fname, open(self.datapath(fname), "rb"))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn("Missing tag cpfDescription in XML file", html)

    def test_export_filename(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            record = testutils.authority_record(cnx, "T-07", "jim")
            for isni, expected_filename in (
                ("", "EAC_{0}.xml".format(record.eid)),
                ("ZZZ/4242", "EAC_ZZZ_4242.xml"),
            ):
                record.cw_set(isni=isni)
                view = self.vreg["views"].select("eac.export", req, record.as_rset())
                view.set_request_content_type()
                self.assertEqual(
                    view._cw.headers_out.getRawHeaders("content-disposition"),
                    ['attachment;filename="{0}"'.format(expected_filename)],
                )

    def test_xmlwrap_component(self):
        with self.admin_access.cnx() as cnx:
            bob = testutils.authority_record(cnx, "T-02", "bob")
            uri = cnx.create_entity("ExternalUri", uri="http://logilab.fr")
            cnx.create_entity(
                "EACResourceRelation",
                resource_relation_agent=bob,
                resource_relation_resource=uri,
                xml_wrap=Binary(b"<plip>plop</plip>"),
            )
            cnx.commit()
        with self.admin_access.web_request() as req:
            rset = req.find("EACResourceRelation")
            component = self.vreg["ctxcomponents"].select(
                "eac.xml_wrap", req, rset=rset
            )
            content = []
            component.render_body(content.append)
        self.assertIn("plop", content[0])
        self.assertEqual(content[0].count("plip"), 2)


class AuthorityRecordFormsTC(CubicWebTC):
    def test_no_copy_action(self):
        with self.admin_access.cnx() as cnx:
            testutils.authority_record(cnx, "T-06", "toto")
            cnx.commit()
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE N name_entry_for X, N parts "toto"')
            actions = self.pactionsdict(req, rset)
            self.assertNotIn(CopyAction, actions["moreactions"])

    def test_unrelated_authorityrecord(self):
        from cubicweb_eac.views import unrelated_authorityrecord

        with self.admin_access.cnx() as cnx:
            ar1 = testutils.authority_record(cnx, "T-1", "1").eid
            ar2 = testutils.authority_record(cnx, "T-2", "2").eid
            # ExternalUri should not appear in choices.
            cnx.create_entity("ExternalUri", uri="http://no-where")
            cnx.commit()
        with self.admin_access.web_request() as req:
            cls = self.vreg["etypes"].etype_class("HierarchicalRelation")(req)
            form = self.vreg["forms"].select("edition", req, entity=cls)
            field = form.field_by_name("hierarchical_parent", "subject")
            choices = unrelated_authorityrecord("hierarchical_parent", form, field)
            expected = [("1", str(ar1)), ("2", str(ar2))]
            self.assertCountEqual(choices, expected)
        # Now with a __linkto.
        linkto = "hierarchical_parent:{0}:subject".format(ar2)
        with self.admin_access.web_request(__linkto=linkto) as req:
            cls = self.vreg["etypes"].etype_class("HierarchicalRelation")(req)
            form = self.vreg["forms"].select("edition", req, entity=cls)
            field = form.field_by_name("hierarchical_parent", "subject")
            choices = unrelated_authorityrecord("hierarchical_parent", form, field)
            expected = [("2", str(ar2))]
            self.assertCountEqual(choices, expected)
            field = form.field_by_name("hierarchical_child", "subject")
            choices = unrelated_authorityrecord("hierarchical_child", form, field)
            expected = [("1", str(ar1)), ("2", str(ar2))]
            self.assertCountEqual(choices, expected)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# copyright 2016-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-eac entity's classes"""

import json

from functools import wraps

from lxml import etree

from logilab.common.date import ustrftime

from cubicweb.predicates import is_instance
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.entity import EntityAdapter

from cubicweb_eac import TYPE_MAPPING, ADDRESS_MAPPING, MAINTENANCETYPE_MAPPING


class DateRelationMixin(object):
    @property
    def start_date(self):
        if self.date_relation:
            return self.date_relation[0].start_date

    @property
    def end_date(self):
        if self.date_relation:
            return self.date_relation[0].end_date


class AuthorityRecord(AnyEntity):
    __regid__ = "AuthorityRecord"
    fetch_attrs, cw_fetch_order = fetch_config(("agent_kind",))

    def dc_title(self):
        """A NameEntry of "authorized" form variant, if any; otherwise any
        NameEntry.
        """
        rset = self._cw.find(
            "NameEntry", name_entry_for=self, form_variant="authorized"
        )
        if not rset:
            rset = self._cw.find("NameEntry", name_entry_for=self)
        return next(rset.entities()).parts

    @property
    def kind(self):
        """The kind of agent"""
        return self.agent_kind[0].name

    @property
    def printable_kind(self):
        """The kind of agent, for display"""
        return self.agent_kind[0].printable_value("name")

    @property
    def other_record_ids(self):
        return sorted(
            [(r.local_type, r.value) for r in self.reverse_eac_other_record_id_of],
            key=lambda x: (x[0] or "", x[1] or ""),
        )


class AgentKind(AnyEntity):
    __regid__ = "AgentKind"
    fetch_attrs, cw_fetch_order = fetch_config(("name",))


class _Relation(DateRelationMixin, AnyEntity):
    __abstract__ = True

    def dc_description(self):
        if self.description:
            return self.description


class ChronologicalRelation(_Relation):
    __regid__ = "ChronologicalRelation"

    @property
    def subject(self):
        return self.chronological_predecessor[0]

    @property
    def object(self):
        return self.chronological_successor[0]


class AssociationRelation(_Relation):
    __regid__ = "AssociationRelation"

    @property
    def subject(self):
        return self.association_from[0]

    @property
    def object(self):
        return self.association_to[0]


class HierarchicalRelation(_Relation):
    __regid__ = "HierarchicalRelation"

    @property
    def subject(self):
        return self.hierarchical_parent[0]

    @property
    def object(self):
        return self.hierarchical_child[0]


class FamilyRelation(_Relation):
    __regid__ = "HierarchicalRelation"

    @property
    def subject(self):
        return self.family_from[0]

    @property
    def object(self):
        return self.family_to[0]


class IdentityRelation(_Relation):
    __regid__ = "HierarchicalRelation"

    @property
    def subject(self):
        return self.identity_from[0]

    @property
    def object(self):
        return self.identity_to[0]


class GeneralContext(AnyEntity):
    __regid__ = "GeneralContext"
    fetch_attrs, cw_fetch_order = fetch_config(("content",))


class JsonAttrsMixin(object):
    @property
    def json_attrs(self):
        return json.loads(self.xml_attributes)


class EACResourceRelation(DateRelationMixin, AnyEntity, JsonAttrsMixin):
    __regid__ = "EACResourceRelation"
    fetch_attrs, cw_fetch_order = fetch_config(
        ("agent_role", "xml_attributes", "resource_role", "description")
    )

    @property
    def record(self):
        return self.resource_relation_agent[0]

    @property
    def resource(self):
        return self.resource_relation_resource[0]

    def dc_title(self):
        agent_title = self.record.dc_title()
        if self.agent_role:
            agent_title += " (%s)" % self.printable_value("agent_role")
        resource_title = self.resource.dc_title()
        if self.resource_role:
            resource_title += " (%s)" % self.printable_value("resource_role")
        return self._cw._("Relation from %(from)s to %(to)s ") % {
            "from": agent_title,
            "to": resource_title,
        }


class EACFunctionRelation(DateRelationMixin, AnyEntity, JsonAttrsMixin):
    __regid__ = "EACFunctionRelation"
    fetch_attrs, cw_fetch_order = fetch_config(
        (
            "r_type",
            "description",
            "relation_entry",
            "place_entry",
            "xml_attributes",
        )
    )

    @property
    def record(self):
        return self.function_relation_agent[0]

    @property
    def resource(self):
        if self.function_relation_function:
            return self.function_relation_function[0]


class DateEntity(AnyEntity):
    __regid__ = "DateEntity"
    fetch_attrs, cw_fetch_order = fetch_config(("start_date", "end_date"))


class SameAsMixIn(object):
    """Mix-in class for entity types supporting vocabulary_source and
    equivalent_concept relations.
    """

    @property
    def scheme(self):
        return self.vocabulary_source and self.vocabulary_source[0] or None

    @property
    def concept(self):
        return self.equivalent_concept and self.equivalent_concept[0] or None


class AgentPlace(DateRelationMixin, SameAsMixIn, AnyEntity):
    __regid__ = "AgentPlace"
    fetch_attrs, cw_fetch_order = fetch_config(("role",))


class AgentFunction(DateRelationMixin, SameAsMixIn, AnyEntity):
    __regid__ = "AgentFunction"
    fetch_attrs, cw_fetch_order = fetch_config(("name", "description"))


class Citation(SameAsMixIn, AnyEntity):
    __regid__ = "Citation"
    fetch_attrs, cw_fetch_order = fetch_config(("uri", "note"))


class Mandate(DateRelationMixin, SameAsMixIn, AnyEntity):
    __regid__ = "Mandate"
    fetch_attrs, cw_fetch_order = fetch_config(("term", "description"))


class LegalStatus(DateRelationMixin, SameAsMixIn, AnyEntity):
    __regid__ = "LegalStatus"


class Occupation(DateRelationMixin, SameAsMixIn, AnyEntity):
    __regid__ = "Occupation"
    fetch_attrs, cw_fetch_order = fetch_config(("term", "description"))


class Structure(AnyEntity):
    __regid__ = "Structure"
    fetch_attrs, cw_fetch_order = fetch_config(("description",))


class EACSource(AnyEntity):
    __regid__ = "EACSource"
    fetch_attrs, cw_fetch_order = fetch_config(("title", "url", "description"))


class EACOtherRecordId(AnyEntity):
    __regid__ = "EACOtherRecordId"
    fetch_attrs, cw_fetch_order = fetch_config(("local_type", "value"))


class History(AnyEntity):
    __regid__ = "History"
    fetch_attrs, cw_fetch_order = fetch_config(("abstract", "text"))


class HistoricalEvent(DateRelationMixin, AnyEntity):
    __regid__ = "HistoricalEvent"
    fetch_attrs, cw_fetch_order = fetch_config(("event",))


class NameEntry(DateRelationMixin, AnyEntity):
    __regid__ = "NameEntry"
    fetch_attrs, cw_fetch_order = fetch_config(
        (
            "language",
            "preferred_form",
            "alternative_form",
            "authorized_form",
            "script_code",
            "parts",
        )
    )


class PlaceEntry(SameAsMixIn, AnyEntity):
    __regid__ = "PlaceEntry"
    fetch_attrs, cw_fetch_order = fetch_config(
        ("name", "latitude", "longitude", "local_type")
    )


class ParallelNames(DateRelationMixin, AnyEntity):
    __regid__ = "ParallelNames"
    fetch_attrs, cw_fetch_order = fetch_config(("authorized_form", "alternative_form"))


# XML export


def substitute_xml_prefix(prefix_name, namespaces):
    """Given an XML prefixed name in the form `'ns:name'`, return the string `'{<ns_uri>}name'`
    where `<ns_uri>` is the URI for the namespace prefix found in `namespaces`.

    This new string is then suitable to build an LXML etree.Element object.

    Example::

        >>> substitude_xml_prefix('xlink:href', {'xlink': 'http://wwww.w3.org/1999/xlink'})
        '{http://www.w3.org/1999/xlink}href'

    """
    try:
        prefix, name = prefix_name.split(":", 1)
    except ValueError:
        return prefix_name
    assert prefix in namespaces, "Unknown namespace prefix: {0}".format(prefix)
    return "{{{0}}}".format(namespaces[prefix]) + name


class AbstractXmlAdapter(EntityAdapter):
    """Abstract adapter to produce XML documents."""

    content_type = "text/xml"
    encoding = "utf-8"
    namespaces = {}

    @property
    def file_name(self):
        """Return a file name for the dump."""
        raise NotImplementedError

    def dump(self):
        """Return an XML string for the adapted entity."""
        raise NotImplementedError

    def element(self, tag, parent=None, attributes=None, text=None):
        """Generic function to build a XSD element tag.

        Params:

        * `name`, value for the 'name' attribute of the xsd:element

        * `parent`, the parent etree node

        * `attributes`, dictionary of attributes
        """
        attributes = attributes or {}
        tag = substitute_xml_prefix(tag, self.namespaces)
        for attr, value in list(attributes.items()):
            newattr = substitute_xml_prefix(attr, self.namespaces)
            attributes[newattr] = value
            if newattr != attr:
                attributes.pop(attr)
        if parent is None:
            elt = etree.Element(tag, attributes, nsmap=self.namespaces)
        else:
            elt = etree.SubElement(parent, tag, attributes)
        if text is not None:
            elt.text = text
        return elt

    @staticmethod
    def cwuri_url(entity):
        """Return an absolute URL for entity's cwuri, necessary for one head ahead application
        handling relative path in cwuri.
        """
        return entity.cwuri


def add_object_xml_wrap(func):
    """Add an objectXMLWrap sub-element to the element returned by `func`."""

    def wrapper(self, entity, *args, **kwargs):
        elem = func(self, entity, *args, **kwargs)
        if entity.xml_wrap is not None:
            objectXMLWrap = self.element("objectXMLWrap", parent=elem)
            objectXMLWrap.append(etree.parse(entity.xml_wrap).getroot())
        return elem

    return wraps(func)(wrapper)


def add_descriptive_note(func):
    @wraps(func)
    def wrapper(self, entity, *args, **kwargs):
        element = func(self, entity, *args, **kwargs)
        description = self._eac_richstring_paragraph_elements(entity, "description")
        if element is not None and description:
            self.element("descriptiveNote", parent=element).extend(description)
        return element

    return wrapper


def add_citation(func):
    @wraps(func)
    def wrapper(self, entity, *args, **kwargs):
        element = func(self, entity, *args, **kwargs)
        for citation in entity.has_citation:
            attrs = {"xlink:type": "simple"}
            if citation.uri:
                attrs["xlink:href"] = citation.uri
            self.element(
                "citation", parent=element, attributes=attrs, text=citation.note
            )
        return element

    return wrapper


class AuthorityRecordEACAdapter(AbstractXmlAdapter):
    __regid__ = "EAC-CPF"
    __select__ = is_instance("AuthorityRecord")

    namespaces = {
        None: "urn:isbn:1-931666-33-4",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "xlink": "http://www.w3.org/1999/xlink",
    }
    datetime_fmt = "%Y-%m-%dT%H:%M:%S"

    @property
    def file_name(self):
        """Return a file name for the dump."""
        if self.entity.isni:
            name = self.entity.isni.replace("/", "_")
        else:
            name = str(self.entity.eid)
        return "EAC_{0}.xml".format(name)

    def dump(self):
        """Return an XML string representing the given agent using the EAC-CPF schema."""
        # Keep related activities since they are used multiple times
        self.activities = sorted(
            self.entity.reverse_generated, key=lambda x: x.start, reverse=True
        )
        # Root element
        eac_cpf_elt = self.element(
            "eac-cpf",
            attributes={
                "xsi:schemaLocation": (
                    "urn:isbn:1-931666-33-4 "
                    "http://eac.staatsbibliothek-berlin.de/schema/cpf.xsd"
                )
            },
        )
        # Top elements: control & cpfDescription
        self.control_element(eac_cpf_elt)
        self.cpfdescription_element(eac_cpf_elt)
        tree = etree.ElementTree(eac_cpf_elt)
        return etree.tostring(
            tree, xml_declaration=True, encoding=self.encoding, pretty_print=True
        )

    def control_element(self, eac_cpf_elt):
        control_elt = self.element("control", parent=eac_cpf_elt)
        self.recordid_element(control_elt)
        self.maintenance_status_element(control_elt)
        self.publication_status_element(control_elt)
        self.maintenance_agency_element(control_elt)
        self.language_declaration_element(control_elt)
        self.maintenance_history_element(control_elt)
        self.sources_element(control_elt)

    def recordid_element(self, control_elt):
        record_id = self.entity.record_id
        if record_id is None:
            record_id = str(self.entity.eid)
        self.element("recordId", parent=control_elt, text=record_id)
        for local_type, value in self.entity.other_record_ids:
            attrs = {}
            if local_type is not None:
                attrs["localType"] = local_type
            self.element(
                "otherRecordId", parent=control_elt, attributes=attrs, text=value
            )

    def cpfdescription_element(self, eac_cpf_elt):
        cpfdescription_elt = self.element("cpfDescription", parent=eac_cpf_elt)
        self.identity_element(cpfdescription_elt)
        self.description_element(cpfdescription_elt)
        self.relations_element(cpfdescription_elt)

    def maintenance_status_element(self, control_elt):
        if any(activity.type == "modify" for activity in self.activities):
            status = "revised"
        else:
            status = "new"
        self.element("maintenanceStatus", parent=control_elt, text=status)

    def publication_status_element(self, control_elt):
        self.element(
            "publicationStatus", parent=control_elt, text="inProcess"
        )  # or approved?

    def maintenance_agency_element(self, control_elt):
        agency_elt = self.element("maintenanceAgency", parent=control_elt)
        self.element("agencyName", parent=agency_elt)

    def language_declaration_element(self, control_elt):
        lang_decl_elt = self.element("languageDeclaration", parent=control_elt)
        self.element(
            "language",
            parent=lang_decl_elt,
            attributes={"languageCode": "fre"},
            text="français",
        )
        self.element(
            "script",
            parent=lang_decl_elt,
            attributes={"scriptCode": "Latn"},
            text="latin",
        )

    def maintenance_history_element(self, control_elt):
        if self.activities:
            history_elt = self.element("maintenanceHistory", parent=control_elt)
            for activity in self.activities:
                self.maintenance_event_element(activity, history_elt)

    def sources_element(self, control_elt):
        sources_elt = self.element("sources")
        for eac_source in self.entity.reverse_source_agent:
            sources_elt.append(self.source_element(eac_source))
        if len(sources_elt):
            control_elt.append(sources_elt)

    def identity_element(self, cpfdescription_elt):
        identity_elt = self.element("identity", parent=cpfdescription_elt)
        self._elt_text_from_attr("entityId", self.entity, "isni", parent=identity_elt)
        type_mapping = dict((v, k) for k, v in TYPE_MAPPING.items())
        eac_type = type_mapping.get(self.entity.kind)
        self.element("entityType", parent=identity_elt, text=eac_type)
        for name_entry in self.entity.reverse_name_entry_for:
            elem = self.element("nameEntry", parent=identity_elt)
            for part in name_entry.parts.split(", "):
                self.element("part", parent=elem, text=part)
            if name_entry.form_variant == "authorized":
                self.element(
                    "authorizedForm", text="conventionDeclaration", parent=elem
                )
            if name_entry.form_variant == "alternative":
                self.element(
                    "alternativeForm", text="conventionDeclaration", parent=elem
                )

    def description_element(self, cpfdescription_elt):
        agent = self.entity
        description_elt = self.element("description", parent=cpfdescription_elt)
        self.exist_dates_element(description_elt)
        for relation, group_tagname, get_element in [
            ("reverse_place_agent", "places", self.place_element),
            ("reverse_function_agent", "functions", self.function_element),
            ("reverse_legal_status_agent", "legalStatuses", self.legal_status_element),
            ("reverse_occupation_agent", "occupations", self.occupation_element),
            ("reverse_mandate_agent", "mandates", self.mandate_element),
        ]:
            relateds = getattr(agent, relation)
            if relateds:
                group_elt = self.element(group_tagname, parent=description_elt)
                for related in relateds:
                    group_elt.append(get_element(related))
        for structure in agent.reverse_structure_agent:
            structure_elt = self.structure_element(structure)
            if len(structure_elt):
                description_elt.append(structure_elt)
        for generalcontext in agent.reverse_general_context_of:
            generalcontext_elt = self.generalcontext_element(generalcontext)
            if len(generalcontext_elt):
                description_elt.append(generalcontext_elt)
        for history in agent.reverse_history_agent:
            bioghist_elt = self.bioghist_element(history)
            if len(bioghist_elt):
                description_elt.append(bioghist_elt)

    def relations_element(self, cpfdescription_elt):
        relations_elt = self.element("relations")
        for rtype, rel_rtype, eac_rtype in [
            (
                "reverse_hierarchical_child",
                "hierarchical_parent",
                "hierarchical-parent",
            ),
            ("reverse_hierarchical_parent", "hierarchical_child", "hierarchical-child"),
            (
                "reverse_chronological_successor",
                "chronological_predecessor",
                "temporal-earlier",
            ),
            (
                "reverse_chronological_predecessor",
                "chronological_successor",
                "temporal-later",
            ),
            ("reverse_association_to", "association_from", "associative"),
            ("reverse_association_from", "association_to", "associative"),
        ]:
            for relation in getattr(self.entity, rtype):
                relations_elt.append(
                    self.cpfrelation_element(relation, rel_rtype, eac_rtype)
                )
        for resource_relation in self.entity.reverse_resource_relation_agent:
            relations_elt.append(self.resource_relation_element(resource_relation))
        if len(relations_elt):
            cpfdescription_elt.append(relations_elt)

    def maintenance_event_element(self, activity, history_elt):
        event_elt = self.element("maintenanceEvent", parent=history_elt)
        type_mapping = dict((v, k) for k, v in MAINTENANCETYPE_MAPPING.items())
        activity_type = type_mapping.get(activity.type, "created")
        self.element("eventType", parent=event_elt, text=activity_type)
        self.element(
            "eventDateTime",
            parent=event_elt,
            attributes={
                "standardDateTime": ustrftime(activity.start, fmt=self.datetime_fmt)
            },
            text=ustrftime(activity.start, fmt=self.datetime_fmt),
        )
        self.agent_element(activity, event_elt)
        self._elt_text_from_attr(
            "eventDescription", activity, "description", parent=event_elt
        )

    @add_descriptive_note
    @add_object_xml_wrap
    def source_element(self, eac_source):
        url = eac_source.url
        attributes = {"xlink:href": url, "xlink:type": "simple"} if url else None
        source_elt = self.element("source", attributes=attributes)
        self.element("sourceEntry", parent=source_elt, text=eac_source.title)
        return source_elt

    def exist_dates_element(self, description_elt):
        date_range = self._eac_date_range_xml_elt(
            self.entity.start_date, self.entity.end_date
        )
        if date_range is not None:
            exist_dates = self.element("existDates", parent=description_elt)
            exist_dates.append(date_range)

    @add_citation
    def place_element(self, place):
        place_elt = self.element("place")
        for attr, eac_name in [("role", "placeRole")]:
            self._elt_text_from_attr(eac_name, place, attr, parent=place_elt)
        for place_entry in place.place_entry_relation:
            place_entry_elt = self.element(
                "placeEntry", parent=place_elt, text=place_entry.name
            )
            if place_entry.equivalent_concept:
                place_entry_elt.attrib["vocabularySource"] = self.cwuri_url(
                    place_entry.equivalent_concept[0]
                )
            for attr, eac_name in (
                ("longitude", "longitude"),
                ("latitude", "latitude"),
                ("local_type", "localType"),
            ):
                value = getattr(place_entry, attr)
                if value:
                    place_entry_elt.attrib[eac_name] = str(value)
        for address in place.place_address:
            self.address_element(address, place_elt)
        return place_elt

    @add_descriptive_note
    @add_citation
    def function_element(self, function):
        function_elt = self.element("function")
        term_elt = self._elt_text_from_attr(
            "term", function, "name", parent=function_elt
        )
        if term_elt is not None and function.equivalent_concept:
            term_elt.attrib["vocabularySource"] = self.cwuri_url(
                function.equivalent_concept[0]
            )
        return function_elt

    @add_descriptive_note
    @add_citation
    def legal_status_element(self, legal_status):
        legal_status_elt = self.element("legalStatus")
        self._elt_text_from_attr("term", legal_status, "term", parent=legal_status_elt)
        return legal_status_elt

    @add_descriptive_note
    @add_citation
    def occupation_element(self, occupation):
        occupation_elt = self.element("occupation")
        term_elt = self._elt_text_from_attr(
            "term", occupation, "term", parent=occupation_elt
        )
        if term_elt is not None and occupation.equivalent_concept:
            term_elt.attrib["vocabularySource"] = self.cwuri_url(
                occupation.equivalent_concept[0]
            )
        self._eac_date_range_xml_elt(
            occupation.start_date, occupation.end_date, parent=occupation_elt
        )
        return occupation_elt

    @add_descriptive_note
    @add_citation
    def mandate_element(self, mandate):
        mandate_elt = self.element("mandate")
        term_elt = self._elt_text_from_attr("term", mandate, "term", parent=mandate_elt)
        if term_elt is not None and mandate.equivalent_concept:
            term_elt.attrib["vocabularySource"] = self.cwuri_url(
                mandate.equivalent_concept[0]
            )
        return mandate_elt

    def structure_element(self, structure):
        structure_elt = self.element("structureOrGenealogy")
        if structure.description:
            structure_elt.extend(
                self._eac_richstring_paragraph_elements(structure, "description")
            )
        return structure_elt

    @add_citation
    def generalcontext_element(self, context):
        context_elt = self.element("generalContext")
        if context.content:
            context_elt.extend(
                self._eac_richstring_paragraph_elements(context, "content")
            )
        return context_elt

    @add_citation
    def bioghist_element(self, history):
        bioghist_elt = self.element("biogHist")
        if history.text:
            bioghist_elt.extend(
                self._eac_richstring_paragraph_elements(history, "text")
            )
        return bioghist_elt

    @add_descriptive_note
    @add_object_xml_wrap
    def cpfrelation_element(self, relation, cw_rtype, eac_rtype):
        related = relation.related(cw_rtype).one()  # exactly one target (schema)
        relation_elt = self.element(
            "cpfRelation",
            attributes={
                "cpfRelationType": eac_rtype,
                "xlink:href": related.cwuri,
                "xlink:type": "simple",
            },
        )
        if relation.entry:
            self.element("relationEntry", parent=relation_elt, text=relation.entry)
        self._eac_date_range_xml_elt(
            getattr(relation, "start_date", None),
            getattr(relation, "end_date", None),
            parent=relation_elt,
        )
        return relation_elt

    @add_object_xml_wrap
    def resource_relation_element(self, resource_relation):
        resource = resource_relation.resource_relation_resource[0]
        uri = resource.uri if resource.cw_etype == "ExternalUri" else resource.cwuri
        attrs = {
            "xlink:href": uri,
            "xlink:type": "simple",
        }
        if resource_relation.resource_role:
            attrs["xlink:role"] = resource_relation.resource_role
        if resource_relation.agent_role:
            attrs["resourceRelationType"] = resource_relation.agent_role
        res_rel_elt = self.element("resourceRelation", attributes=attrs)
        self._eac_date_range_xml_elt(
            resource_relation.start_date, resource_relation.end_date, parent=res_rel_elt
        )
        return res_rel_elt

    def agent_element(self, activity, maintenance_event_elt):
        if activity.agent:
            self.element("agentType", maintenance_event_elt, text="human")
            self.element("agent", maintenance_event_elt, text=activity.agent)
        else:  # These tags must be present, even if name is empty
            agent_type_elt = self.element("agentType", text="machine")
            maintenance_event_elt.append(agent_type_elt)
            maintenance_event_elt.append(self.element("agent"))

    def address_element(self, address, place_elt):
        address_elt = self.element("address", parent=place_elt)
        for eac_name, attr in ADDRESS_MAPPING:
            self._elt_text_from_attr(
                "addressLine",
                address,
                attr,
                attributes={"localType": eac_name},
                parent=address_elt,
            )

    # helper methods for lxml

    def _elt_text_from_attr(
        self, tag_name, entity, attr_name, parent=None, attributes=None
    ):
        """Return an lxml `Element` whose text is the value of the given attribute on the given
        entity.

        If this element is not empty and if ``parent`` is not ``None``, the element will also be
        inserted in the parent XML element.

        If ``attributes`` is not ``None``, these attributes will be added to the returned element.

        Return ``None`` if element is empty.
        """
        value = getattr(entity, attr_name)
        if value is not None:
            elt = self.element(
                tag_name, parent=parent, attributes=attributes, text=value
            )
            return elt

    def _eac_date_range_xml_elt(self, start_date, end_date, parent=None):
        """Return an EAC lxml ``'dateRange'`` ``Element`` with the given boundaries."""
        if not start_date and not end_date:
            return
        date_range = self.element("dateRange", parent=parent)
        for dt, eac_name in [(start_date, "fromDate"), (end_date, "toDate")]:
            if not dt:
                continue
            self.element(
                eac_name,
                parent=date_range,
                attributes={"standardDate": dt.isoformat()},
                text=dt.isoformat(),
            )
        return date_range

    def _eac_richstring_paragraph_elements(self, entity, attr_name):
        fmt = getattr(entity, attr_name + "_format")
        if fmt == "text/plain":
            value = getattr(entity, attr_name)
            if not value:
                return []
            return [self.element("p", text=line) for line in value.splitlines()]
        else:
            value = entity.printable_value(attr_name)
            if not value:
                return []
            return list(etree.fromstring("<root>{0}</root>".format(value)))
